"""CLI commands for scheduling history."""

from __future__ import annotations

import sys
from collections.abc import Callable, Sequence
from typing import TYPE_CHECKING, Any, Protocol

import click

from .extensions import pass_ctx_obj
from .types import CLIContext

if TYPE_CHECKING:
    from uuid import UUID

    from ai.backend.common.dto.manager.scheduling_history import (
        DeploymentHistoryDTO,
        RouteHistoryDTO,
        SessionHistoryDTO,
        SubStepResultDTO,
    )

    HistoryDTO = SessionHistoryDTO | DeploymentHistoryDTO | RouteHistoryDTO


class HistoryRecord(Protocol):
    """Protocol for history record types."""

    phase: str
    from_status: str | None
    to_status: str | None
    result: str
    error_code: str | None
    message: str | None
    sub_steps: list[SubStepResultDTO]
    created_at: str


def _format_local_time(utc_timestamp: str | None) -> str:
    """Convert UTC timestamp to local time string.

    Args:
        utc_timestamp: UTC timestamp string in ISO format

    Returns:
        Formatted local time string (YYYY-MM-DD HH:MM:SS)
    """
    if not utc_timestamp:
        return ""
    try:
        from datetime import UTC, datetime

        # Handle ISO format with or without timezone
        ts_str = utc_timestamp
        if ts_str.endswith("Z"):
            ts_str = ts_str[:-1] + "+00:00"
        dt = datetime.fromisoformat(ts_str)

        # If naive datetime, assume UTC
        if dt.tzinfo is None:
            dt = dt.replace(tzinfo=UTC)

        # Convert to local time
        local_dt = dt.astimezone()
        return local_dt.strftime("%Y-%m-%d %H:%M:%S")
    except Exception:
        # Fallback to string slice if parsing fails
        return utc_timestamp[:19]


def _render_sub_steps(sub_steps: list[Any], prefix: str, *, verbose: bool = False) -> None:
    """Render sub_steps with tree structure."""
    for k, step in enumerate(sub_steps):
        is_last_step = k == len(sub_steps) - 1
        step_prefix = f"{prefix}└─" if is_last_step else f"{prefix}├─"

        # Step result indicator with color
        is_success = step.result == "SUCCESS"
        step_icon = click.style("✓", fg="green") if is_success else click.style("✗", fg="red")
        step_name = (
            click.style(step.step, fg="green") if is_success else click.style(step.step, fg="red")
        )

        # Duration
        duration_str = ""
        if step.started_at and step.ended_at:
            try:
                from datetime import datetime

                start = datetime.fromisoformat(str(step.started_at))
                end = datetime.fromisoformat(str(step.ended_at))
                duration = (end - start).total_seconds()
                duration_str = f" ({duration:.2f}s)"
            except Exception:
                pass

        step_line = f"{step_prefix} {step_icon} {step_name}{duration_str}"

        # Add error message for failed steps
        if step.result != "SUCCESS" and step.message:
            step_msg = (
                step.message if verbose or len(step.message) <= 40 else step.message[:37] + "..."
            )
            step_line += f': "{step_msg}"'

        print(step_line)


def _render_flat_view(
    entity_type: str,
    histories: Sequence[HistoryDTO],
    get_entity_label: Callable[[Any], str],
    *,
    verbose: bool = False,
) -> None:
    """Render history records in flat chronological order.

    Args:
        entity_type: Type name (e.g., "Session", "Deployment", "Route")
        histories: List of history records (already sorted by time)
        get_entity_label: Function to get display label for entity
        verbose: If True, show full messages without truncation
    """
    if not histories:
        print(f"No {entity_type.lower()} history found")
        return

    print(f"{entity_type} History ({len(histories)} records)")
    print("│")

    for i, h in enumerate(histories):
        is_last = i == len(histories) - 1
        main_prefix = "└─" if is_last else "├─"
        detail_prefix = "   " if is_last else "│  "

        # Result indicator with color
        is_success = h.result == "SUCCESS"
        if is_success:
            result_indicator = click.style("[SUCCESS]", fg="green")
            phase_str = click.style(h.phase, fg="green")
        else:
            result_indicator = click.style("[FAILURE]", fg="red")
            phase_str = click.style(h.phase, fg="red")

        # Entity label
        entity_label = get_entity_label(h)

        # Status transition
        status_str = ""
        if h.from_status or h.to_status:
            status_str = f" ({h.from_status or '?'} → {h.to_status or '?'})"

        # Format timestamp (convert UTC to local time)
        created = _format_local_time(str(h.created_at) if h.created_at else None)

        print(
            f"{main_prefix} {result_indicator} {entity_label} | {phase_str}{status_str} @ {created}"
        )

        # Show error info if failure
        has_sub_steps = bool(h.sub_steps)
        if h.result != "SUCCESS" and (h.error_code or h.message):
            if h.error_code:
                error_text = click.style(f"Error: {h.error_code}", fg="red")
                connector = "│  " if (h.message or has_sub_steps) else "   "
                print(f"{detail_prefix}{connector}{error_text}")
            if h.message:
                msg = h.message if verbose or len(h.message) <= 60 else h.message[:57] + "..."
                connector = "│  " if has_sub_steps else "   "
                print(f"{detail_prefix}{connector}Message: {msg}")

        # Show sub_steps
        if has_sub_steps:
            _render_sub_steps(h.sub_steps, detail_prefix, verbose=verbose)

        if not is_last:
            print("│")


def _render_grouped_view(
    entity_type: str,
    histories: Sequence[HistoryDTO],
    get_entity_id: Callable[[Any], UUID],
    get_entity_label: Callable[[Any], str],
    *,
    verbose: bool = False,
) -> None:
    """Render history records grouped by entity.

    Args:
        entity_type: Type name (e.g., "Session", "Deployment", "Route")
        histories: List of history records
        get_entity_id: Function to extract entity ID from history record
        get_entity_label: Function to get display label for entity
        verbose: If True, show full messages without truncation
    """
    if not histories:
        print(f"No {entity_type.lower()} history found")
        return

    # Group by entity ID
    grouped: dict[str, list[HistoryDTO]] = {}
    entity_labels: dict[str, str] = {}
    for h in histories:
        eid = str(get_entity_id(h))
        if eid not in grouped:
            grouped[eid] = []
            entity_labels[eid] = get_entity_label(h)
        grouped[eid].append(h)

    print(f"{entity_type} History ({len(histories)} records)")
    print("│")

    entity_ids = list(grouped.keys())
    for i, entity_id in enumerate(entity_ids):
        is_last_entity = i == len(entity_ids) - 1
        entity_prefix = "└─" if is_last_entity else "├─"
        child_prefix = "   " if is_last_entity else "│  "

        print(f"{entity_prefix} {entity_labels[entity_id]}")

        entity_histories = grouped[entity_id]
        for j, h in enumerate(entity_histories):
            is_last_history = j == len(entity_histories) - 1
            history_prefix = f"{child_prefix}└─" if is_last_history else f"{child_prefix}├─"
            detail_prefix = f"{child_prefix}   " if is_last_history else f"{child_prefix}│  "

            # Result indicator with color
            is_success = h.result == "SUCCESS"
            if is_success:
                result_indicator = click.style("[SUCCESS]", fg="green")
                phase_str = click.style(h.phase, fg="green")
            else:
                result_indicator = click.style("[FAILURE]", fg="red")
                phase_str = click.style(h.phase, fg="red")

            # Status transition
            status_str = ""
            if h.from_status or h.to_status:
                status_str = f" ({h.from_status or '?'} → {h.to_status or '?'})"

            # Format timestamp (convert UTC to local time)
            created = _format_local_time(str(h.created_at) if h.created_at else None)

            print(f"{history_prefix} {result_indicator} {phase_str}{status_str} @ {created}")

            # Show error info if failure
            has_sub_steps = bool(h.sub_steps)
            if h.result != "SUCCESS" and (h.error_code or h.message):
                if h.error_code:
                    error_text = click.style(f"Error: {h.error_code}", fg="red")
                    connector = "│  " if (h.message or has_sub_steps) else "   "
                    print(f"{detail_prefix}{connector}{error_text}")
                if h.message:
                    msg = h.message if verbose or len(h.message) <= 60 else h.message[:57] + "..."
                    connector = "│  " if has_sub_steps else "   "
                    print(f"{detail_prefix}{connector}Message: {msg}")

            # Show sub_steps
            if has_sub_steps:
                _render_sub_steps(h.sub_steps, detail_prefix, verbose=verbose)

        if not is_last_entity:
            print("│")


@click.group()
def scheduling_history() -> None:
    """Scheduling history operations (superadmin only)"""


# Session scheduling history


@scheduling_history.group()
def session() -> None:
    """View session scheduling history"""


@session.command("list")
@pass_ctx_obj
@click.option("--session-id", type=str, default=None, help="Filter by session ID")
@click.option("--phase", type=str, default=None, help="Filter by phase (contains)")
@click.option(
    "--from-status",
    multiple=True,
    default=None,
    help="Filter by from_status (can specify multiple)",
)
@click.option(
    "--to-status", multiple=True, default=None, help="Filter by to_status (can specify multiple)"
)
@click.option(
    "--result",
    type=click.Choice(["SUCCESS", "FAILURE", "STALE"]),
    default=None,
    help="Filter by scheduling result",
)
@click.option("--error-code", type=str, default=None, help="Filter by error code")
@click.option("--message", type=str, default=None, help="Filter by message (contains)")
@click.option("--limit", type=int, default=20, help="Maximum number of records to return")
@click.option("--offset", type=int, default=0, help="Offset for pagination")
@click.option(
    "--order-by",
    type=click.Choice(["created_at", "updated_at"]),
    default="created_at",
    help="Order by field",
)
@click.option(
    "--order",
    type=click.Choice(["ASC", "DESC"]),
    default="DESC",
    help="Order direction (default: DESC to get latest records)",
)
@click.option("--json", "as_json", is_flag=True, help="Output as JSON")
@click.option("--group", "group_by_entity", is_flag=True, help="Group by session ID")
@click.option("-v", "--verbose", is_flag=True, help="Show full messages without truncation")
def list_session_history_cmd(
    ctx: CLIContext,
    session_id: str | None,
    phase: str | None,
    from_status: tuple[str, ...],
    to_status: tuple[str, ...],
    result: str | None,
    error_code: str | None,
    message: str | None,
    limit: int,
    offset: int,
    order_by: str,
    order: str,
    as_json: bool,
    group_by_entity: bool,
    verbose: bool,
) -> None:
    """
    List session scheduling history records.
    """
    import json
    from uuid import UUID

    from ai.backend.cli.types import ExitCode
    from ai.backend.client.session import Session
    from ai.backend.common.dto.manager.query import StringFilter, UUIDFilter
    from ai.backend.common.dto.manager.scheduling_history import (
        OrderDirection,
        SchedulingResultType,
        SearchSessionHistoryRequest,
        SessionHistoryFilter,
        SessionHistoryOrder,
        SessionHistoryOrderField,
    )

    with Session() as api_session:
        try:
            # Build filter if any filter options are provided
            has_filter = any([
                session_id,
                phase,
                from_status,
                to_status,
                result,
                error_code,
                message,
            ])
            filter_cond = None
            if has_filter:
                filter_cond = SessionHistoryFilter(
                    session_id=UUIDFilter(equals=UUID(session_id)) if session_id else None,
                    phase=StringFilter(contains=phase) if phase else None,
                    from_status=list(from_status) if from_status else None,
                    to_status=list(to_status) if to_status else None,
                    result=[SchedulingResultType(result)] if result else None,
                    error_code=StringFilter(contains=error_code) if error_code else None,
                    message=StringFilter(contains=message) if message else None,
                )

            order_spec = SessionHistoryOrder(
                field=SessionHistoryOrderField(order_by),
                direction=OrderDirection(order.lower()),
            )

            request = SearchSessionHistoryRequest(
                filter=filter_cond,
                order=[order_spec],
                limit=limit,
                offset=offset,
            )
            response = api_session.SchedulingHistory.list_session_history(request)

            # Reverse for display when DESC (newest at bottom for readability)
            display_items = (
                list(reversed(response.items)) if order == "DESC" else list(response.items)
            )

            if as_json:
                print(
                    json.dumps(
                        [h.model_dump(mode="json") for h in response.items],
                        indent=2,
                        default=str,
                    )
                )
            elif group_by_entity:
                _render_grouped_view(
                    "Session",
                    display_items,
                    lambda h: h.session_id,
                    lambda h: f"Session: {h.session_id}",
                    verbose=verbose,
                )
            else:
                _render_flat_view(
                    "Session",
                    display_items,
                    lambda h: f"Session: {h.session_id}",
                    verbose=verbose,
                )
        except Exception as e:
            ctx.output.print_error(e)
            sys.exit(ExitCode.FAILURE)


# Deployment history


@scheduling_history.group()
def deployment() -> None:
    """View deployment scheduling history"""


@deployment.command("list")
@pass_ctx_obj
@click.option("--deployment-id", type=str, default=None, help="Filter by deployment ID")
@click.option("--phase", type=str, default=None, help="Filter by phase (contains)")
@click.option(
    "--from-status",
    multiple=True,
    default=None,
    help="Filter by from_status (can specify multiple)",
)
@click.option(
    "--to-status", multiple=True, default=None, help="Filter by to_status (can specify multiple)"
)
@click.option(
    "--result",
    type=click.Choice(["SUCCESS", "FAILURE", "STALE"]),
    default=None,
    help="Filter by scheduling result",
)
@click.option("--error-code", type=str, default=None, help="Filter by error code")
@click.option("--message", type=str, default=None, help="Filter by message (contains)")
@click.option("--limit", type=int, default=20, help="Maximum number of records to return")
@click.option("--offset", type=int, default=0, help="Offset for pagination")
@click.option(
    "--order-by",
    type=click.Choice(["created_at", "updated_at"]),
    default="created_at",
    help="Order by field",
)
@click.option(
    "--order",
    type=click.Choice(["ASC", "DESC"]),
    default="DESC",
    help="Order direction (default: DESC to get latest records)",
)
@click.option("--json", "as_json", is_flag=True, help="Output as JSON")
@click.option("--group", "group_by_entity", is_flag=True, help="Group by deployment ID")
@click.option("-v", "--verbose", is_flag=True, help="Show full messages without truncation")
def list_deployment_history_cmd(
    ctx: CLIContext,
    deployment_id: str | None,
    phase: str | None,
    from_status: tuple[str, ...],
    to_status: tuple[str, ...],
    result: str | None,
    error_code: str | None,
    message: str | None,
    limit: int,
    offset: int,
    order_by: str,
    order: str,
    as_json: bool,
    group_by_entity: bool,
    verbose: bool,
) -> None:
    """
    List deployment scheduling history records.
    """
    import json
    from uuid import UUID

    from ai.backend.cli.types import ExitCode
    from ai.backend.client.session import Session
    from ai.backend.common.dto.manager.query import StringFilter, UUIDFilter
    from ai.backend.common.dto.manager.scheduling_history import (
        DeploymentHistoryFilter,
        DeploymentHistoryOrder,
        DeploymentHistoryOrderField,
        OrderDirection,
        SchedulingResultType,
        SearchDeploymentHistoryRequest,
    )

    with Session() as api_session:
        try:
            # Build filter if any filter options are provided
            has_filter = any([
                deployment_id,
                phase,
                from_status,
                to_status,
                result,
                error_code,
                message,
            ])
            filter_cond = None
            if has_filter:
                filter_cond = DeploymentHistoryFilter(
                    deployment_id=UUIDFilter(equals=UUID(deployment_id)) if deployment_id else None,
                    phase=StringFilter(contains=phase) if phase else None,
                    from_status=list(from_status) if from_status else None,
                    to_status=list(to_status) if to_status else None,
                    result=[SchedulingResultType(result)] if result else None,
                    error_code=StringFilter(contains=error_code) if error_code else None,
                    message=StringFilter(contains=message) if message else None,
                )

            order_spec = DeploymentHistoryOrder(
                field=DeploymentHistoryOrderField(order_by),
                direction=OrderDirection(order.lower()),
            )

            request = SearchDeploymentHistoryRequest(
                filter=filter_cond,
                order=[order_spec],
                limit=limit,
                offset=offset,
            )
            response = api_session.SchedulingHistory.list_deployment_history(request)

            # Reverse for display when DESC (newest at bottom for readability)
            display_items = (
                list(reversed(response.items)) if order == "DESC" else list(response.items)
            )

            if as_json:
                print(
                    json.dumps(
                        [h.model_dump(mode="json") for h in response.items],
                        indent=2,
                        default=str,
                    )
                )
            elif group_by_entity:
                _render_grouped_view(
                    "Deployment",
                    display_items,
                    lambda h: h.deployment_id,
                    lambda h: f"Deployment: {h.deployment_id}",
                    verbose=verbose,
                )
            else:
                _render_flat_view(
                    "Deployment",
                    display_items,
                    lambda h: f"Deployment: {h.deployment_id}",
                    verbose=verbose,
                )
        except Exception as e:
            ctx.output.print_error(e)
            sys.exit(ExitCode.FAILURE)


# Route history


@scheduling_history.group()
def route() -> None:
    """View route scheduling history"""


@route.command("list")
@pass_ctx_obj
@click.option("--route-id", type=str, default=None, help="Filter by route ID")
@click.option("--deployment-id", type=str, default=None, help="Filter by deployment ID")
@click.option("--phase", type=str, default=None, help="Filter by phase (contains)")
@click.option(
    "--from-status",
    multiple=True,
    default=None,
    help="Filter by from_status (can specify multiple)",
)
@click.option(
    "--to-status", multiple=True, default=None, help="Filter by to_status (can specify multiple)"
)
@click.option(
    "--result",
    type=click.Choice(["SUCCESS", "FAILURE", "STALE"]),
    default=None,
    help="Filter by scheduling result",
)
@click.option("--error-code", type=str, default=None, help="Filter by error code")
@click.option("--message", type=str, default=None, help="Filter by message (contains)")
@click.option("--limit", type=int, default=20, help="Maximum number of records to return")
@click.option("--offset", type=int, default=0, help="Offset for pagination")
@click.option(
    "--order-by",
    type=click.Choice(["created_at", "updated_at"]),
    default="created_at",
    help="Order by field",
)
@click.option(
    "--order",
    type=click.Choice(["ASC", "DESC"]),
    default="DESC",
    help="Order direction (default: DESC to get latest records)",
)
@click.option("--json", "as_json", is_flag=True, help="Output as JSON")
@click.option("--group", "group_by_entity", is_flag=True, help="Group by route ID")
@click.option("-v", "--verbose", is_flag=True, help="Show full messages without truncation")
def list_route_history_cmd(
    ctx: CLIContext,
    route_id: str | None,
    deployment_id: str | None,
    phase: str | None,
    from_status: tuple[str, ...],
    to_status: tuple[str, ...],
    result: str | None,
    error_code: str | None,
    message: str | None,
    limit: int,
    offset: int,
    order_by: str,
    order: str,
    as_json: bool,
    group_by_entity: bool,
    verbose: bool,
) -> None:
    """
    List route scheduling history records.
    """
    import json
    from uuid import UUID

    from ai.backend.cli.types import ExitCode
    from ai.backend.client.session import Session
    from ai.backend.common.dto.manager.query import StringFilter, UUIDFilter
    from ai.backend.common.dto.manager.scheduling_history import (
        OrderDirection,
        RouteHistoryFilter,
        RouteHistoryOrder,
        RouteHistoryOrderField,
        SchedulingResultType,
        SearchRouteHistoryRequest,
    )

    with Session() as api_session:
        try:
            # Build filter if any filter options are provided
            has_filter = any([
                route_id,
                deployment_id,
                phase,
                from_status,
                to_status,
                result,
                error_code,
                message,
            ])
            filter_cond = None
            if has_filter:
                filter_cond = RouteHistoryFilter(
                    route_id=UUIDFilter(equals=UUID(route_id)) if route_id else None,
                    deployment_id=UUIDFilter(equals=UUID(deployment_id)) if deployment_id else None,
                    phase=StringFilter(contains=phase) if phase else None,
                    from_status=list(from_status) if from_status else None,
                    to_status=list(to_status) if to_status else None,
                    result=[SchedulingResultType(result)] if result else None,
                    error_code=StringFilter(contains=error_code) if error_code else None,
                    message=StringFilter(contains=message) if message else None,
                )

            order_spec = RouteHistoryOrder(
                field=RouteHistoryOrderField(order_by),
                direction=OrderDirection(order.lower()),
            )

            request = SearchRouteHistoryRequest(
                filter=filter_cond,
                order=[order_spec],
                limit=limit,
                offset=offset,
            )
            response = api_session.SchedulingHistory.list_route_history(request)

            # Reverse for display when DESC (newest at bottom for readability)
            display_items = (
                list(reversed(response.items)) if order == "DESC" else list(response.items)
            )

            def get_route_label(h: Any) -> str:
                return f"Route: {h.route_id} (Deployment: {h.deployment_id})"

            if as_json:
                print(
                    json.dumps(
                        [h.model_dump(mode="json") for h in response.items],
                        indent=2,
                        default=str,
                    )
                )
            elif group_by_entity:
                _render_grouped_view(
                    "Route",
                    display_items,
                    lambda h: h.route_id,
                    get_route_label,
                    verbose=verbose,
                )
            else:
                _render_flat_view(
                    "Route",
                    display_items,
                    get_route_label,
                    verbose=verbose,
                )
        except Exception as e:
            ctx.output.print_error(e)
            sys.exit(ExitCode.FAILURE)



from __future__ import annotations

import builtins
import sys
import uuid
from collections.abc import Callable
from typing import Any

import click

from ai.backend.cli.main import main
from ai.backend.cli.types import ExitCode
from ai.backend.client.cli.extensions import pass_ctx_obj
from ai.backend.client.cli.pretty import print_fail
from ai.backend.client.cli.session.lifecycle import session as user_session
from ai.backend.client.cli.types import CLIContext
from ai.backend.client.output.fields import session_fields, session_fields_v5
from ai.backend.client.output.types import FieldSpec
from ai.backend.client.session import Session

from . import admin

SessionItem = dict[str, Any]


@admin.group()
def session() -> None:
    """
    Session administration commands.
    """


def _list_cmd(name: str = "list", docs: str | None = None) -> Callable[..., None]:
    @pass_ctx_obj
    @click.option(
        "-s",
        "--status",
        default=None,
        type=click.Choice([
            "PENDING",
            "SCHEDULED",
            "PULLING",
            "PREPARED",
            "PREPARING",
            "CREATING",
            "RUNNING",
            "RESTARTING",
            "RUNNING_DEGRADED",
            "TERMINATING",
            "TERMINATED",
            "ERROR",
            "CANCELLED",
            "ALL",  # special case
        ]),
        help="Filter by the given status",
    )
    @click.option(
        "--access-key",
        type=str,
        default=None,
        help="Get sessions for a specific access key (only works if you are a super-admin)",
    )
    @click.option("--name-only", is_flag=True, help="Display session names only.")
    @click.option(
        "--dead", is_flag=True, help="Filter only dead sessions. Ignores --status option."
    )
    @click.option(
        "--running",
        is_flag=True,
        help="Filter only scheduled and running sessions. Ignores --status option.",
    )
    @click.option("--detail", is_flag=True, help="Show more details using more columns.")
    @click.option("-f", "--format", default=None, help="Display only specified fields.")
    @click.option(
        "--plain",
        is_flag=True,
        help="Display the session list without decorative line drawings and the header.",
    )
    @click.option("--filter", "filter_", default=None, help="Set the query filter expression.")
    @click.option("--order", default=None, help="Set the query ordering expression.")
    @click.option(
        "--offset", default=0, type=int, help="The index of the current page start for pagination."
    )
    @click.option("--limit", default=None, type=int, help="The page size for pagination.")
    @click.option(
        "-a",
        "--all",
        is_flag=True,
        default=False,
        help=(
            'Alias of "backend.ai ps --status=ALL" listing all sessions regardless of status.'
            " Ignores --status option."
        ),
    )
    def list(
        ctx: CLIContext,
        status: str | None,
        access_key: str | None,
        name_only: str | None,
        dead: bool,
        running: bool,
        detail: bool,
        format: str | None,
        plain: bool,
        filter_: str | None,
        order: str | None,
        offset: int,
        limit: int | None,
        all: bool,
    ) -> None:
        """
        List and manage compute sessions.
        """
        fields: builtins.list[FieldSpec] = []
        with Session() as api_sess:
            is_admin = api_sess.KeyPair(api_sess.config.access_key).info()["is_admin"]
            api_version = api_sess.api_version
            try:
                fields.append(session_fields["name"])
                if is_admin:
                    fields.append(session_fields["access_key"])
            except Exception as e:
                ctx.output.print_error(e)
                sys.exit(ExitCode.FAILURE)
            if name_only:
                pass
            elif format is not None:
                options = format.split(",")
                for opt in options:
                    if opt not in session_fields or opt == "containers":
                        ctx.output.print_fail(f"There is no such format option: {opt}")
                        sys.exit(ExitCode.INVALID_ARGUMENT)
                fields = [session_fields[opt] for opt in options]
            else:
                if api_version >= (8, "20240915"):
                    fields.append(session_fields["priority"])
                if api_version[0] >= 6:
                    fields.append(session_fields["session_id"])
                fields.extend([
                    session_fields["group_name"],
                    session_fields["type"],
                    session_fields["image"],
                    session_fields["status"],
                    session_fields["status_info"],
                    session_fields["status_changed"],
                    session_fields["result"],
                ])
                if detail:
                    fields.extend([
                        session_fields["tag"],
                        session_fields["created_at"],
                        session_fields["occupying_slots"],
                    ])

        no_match_name = None
        if status is None:
            status = ",".join([
                "PENDING",
                "SCHEDULED",
                "PULLING",
                "PREPARED",
                "PREPARING",
                "CREATING",
                "RUNNING",
                "RUNNING_DEGRADED",
                "TERMINATING",
                "ERROR",
            ])
            no_match_name = "active"
        if running:
            status = ",".join([
                "CREATING",
                "RUNNING",
                "RUNNING_DEGRADED",
            ])
            no_match_name = "running"
        if dead:
            status = ",".join([
                "CANCELLED",
                "TERMINATED",
            ])
            no_match_name = "dead"
        if status == "ALL" or all:
            status = ",".join([
                "PENDING",
                "SCHEDULED",
                "PULLING",
                "PREPARED",
                "PREPARING",
                "CREATING",
                "RUNNING",
                "RESTARTING",
                "RUNNING_DEGRADED",
                "TERMINATING",
                "TERMINATED",
                "ERROR",
                "CANCELLED",
            ])
            no_match_name = "in any status"
        if no_match_name is None:
            no_match_name = status.lower()

        try:
            with Session() as api_sess:
                fetch_func = lambda pg_offset, pg_size: api_sess.ComputeSession.paginated_list(
                    status,
                    access_key,
                    fields=fields,
                    page_offset=pg_offset,
                    page_size=pg_size,
                    filter=filter_,
                    order=order,
                )
                ctx.output.print_paginated_list(
                    fetch_func,
                    initial_page_offset=offset,
                    page_size=limit,
                    plain=plain,
                )
        except Exception as e:
            ctx.output.print_error(e)
            sys.exit(ExitCode.FAILURE)

    list.__name__ = name
    if docs is not None:
        list.__doc__ = docs
    return list


# Make it available as:
# - backend.ai ps
# - backend.ai admin session list
main.command()(_list_cmd(name="ps", docs='Alias of "session list"'))
user_session.command()(_list_cmd(docs='Alias of "admin session list"'))
session.command()(_list_cmd())


def _info_cmd(docs: str | None = None) -> Callable[..., None]:
    @pass_ctx_obj
    @click.argument("session_id", metavar="SESSID")
    def info(ctx: CLIContext, session_id: str) -> None:
        """
        Show detailed information for a running compute session.
        """
        with Session() as api_sess:
            api_version = api_sess.api_version
            fields = [
                session_fields["name"],
            ]
            if api_version >= (8, "20240915"):
                fields.append(session_fields["priority"])
            if api_version[0] >= 6:
                fields.append(session_fields["session_id"])
                fields.append(session_fields["main_kernel_id"])
            fields.extend([
                session_fields["image"],
                session_fields["tag"],
                session_fields["created_at"],
                session_fields["terminated_at"],
                session_fields["status"],
                session_fields["status_info"],
                session_fields["status_data"],
                session_fields["occupying_slots"],
                session_fields["idle_checks"],
            ])
            if api_sess.api_version[0] >= 6:
                fields.append(session_fields["containers"])
            else:
                fields.append(session_fields_v5["containers"])
            fields.append(session_fields["dependencies"])
            q = "query($id: UUID!) {  compute_session(id: $id) {    $fields  }}"
            try:
                uuid.UUID(session_id)
            except ValueError:
                print_fail("In API v5 or later, the session ID must be given in the UUID format.")
                sys.exit(ExitCode.FAILURE)
            v = {"id": session_id}
            q = q.replace("$fields", " ".join(f.field_ref for f in fields))
            try:
                resp = api_sess.Admin.query(q, v)
            except Exception as e:
                ctx.output.print_error(e)
                sys.exit(ExitCode.FAILURE)
            if resp["compute_session"] is None:
                if api_sess.api_version[0] < 5:
                    ctx.output.print_fail("There is no such running compute session.")
                else:
                    ctx.output.print_fail("There is no such compute session.")
                sys.exit(ExitCode.FAILURE)
            ctx.output.print_item(resp["compute_session"], fields)

    if docs is not None:
        info.__doc__ = docs
    return info


main.command()(_info_cmd(docs='Alias of "session info"'))
user_session.command()(_info_cmd(docs='Alias of "admin session info"'))
session.command()(_info_cmd())



from __future__ import annotations

import json
from collections.abc import Callable, Mapping, Sequence
from typing import Any, TypeVar, cast

from ai.backend.client.exceptions import BackendAPIError
from ai.backend.common.types import ResultSet

from .types import BaseOutputHandler, FieldSpec, PaginatedResult

_json_opts: Mapping[str, Any] = {"indent": 2}

T = TypeVar("T")


class JsonOutputHandler(BaseOutputHandler):
    def print_item(
        self,
        item: Mapping[str, Any] | None,
        fields: Sequence[FieldSpec],
    ) -> None:
        if item is None:
            print(
                json.dumps({
                    "count": 0,
                    "total_count": 0,
                    "items": [],
                })
            )
            return
        field_map = {f.field_name: f for f in fields}
        print(
            json.dumps(
                {
                    "count": 1,
                    "total_count": 1,
                    "items": [
                        {
                            field_map[k].alt_name: field_map[k].formatter.format_json(
                                v, field_map[k]
                            )
                            for k, v in item.items()
                            if k in field_map
                        },
                    ],
                },
                **_json_opts,
            )
        )

    def print_items(
        self,
        items: Sequence[Mapping[str, Any]],
        fields: Sequence[FieldSpec],
    ) -> None:
        field_map = {f.field_name: f for f in fields}
        print(
            json.dumps(
                {
                    "count": len(items),
                    "total_count": len(items),
                    "items": [
                        {
                            field_map[k].alt_name: field_map[k].formatter.format_json(
                                v, field_map[k]
                            )
                            for k, v in item.items()
                            if k in field_map
                        }
                        for item in items
                    ],
                },
                **_json_opts,
            )
        )

    def print_result_set(
        self,
        result_set: ResultSet,
    ) -> None:
        print(json.dumps(result_set))

    def print_list(
        self,
        items: Sequence[Mapping[str, Any]],
        fields: Sequence[FieldSpec],
        *,
        is_scalar: bool = False,
    ) -> None:
        if is_scalar:
            if len(fields) != 1:
                raise ValueError("Scalar output requires exactly one field")
            item_list = [
                {
                    fields[0].alt_name: fields[0].formatter.format_json(item, fields[0]),
                }
                for item in items
            ]
        else:
            field_map = {f.field_name: f for f in fields}
            item_list = [
                {
                    field_map[k].alt_name: field_map[k].formatter.format_json(v, field_map[k])
                    for k, v in item.items()
                    if k in field_map
                }
                for item in items
            ]
        print(
            json.dumps(
                {
                    "count": len(items),
                    "total_count": len(items),
                    "items": item_list,
                },
                **_json_opts,
            )
        )

    def print_paginated_list(
        self,
        fetch_func: Callable[[int, int], PaginatedResult[T]],
        initial_page_offset: int,
        page_size: int | None = None,
        plain: bool = False,
    ) -> None:
        page_size = page_size or 20
        result = fetch_func(initial_page_offset, page_size)
        field_map = {f.field_name: f for f in result.fields}
        print(
            json.dumps(
                {
                    "count": len(result.items),
                    "total_count": result.total_count,
                    "items": [
                        {
                            field_map[k].alt_name: field_map[k].formatter.format_json(
                                v, field_map[k]
                            )
                            for k, v in cast(Mapping[str, Any], item).items()
                            if k in field_map
                        }
                        for item in result.items
                    ],
                },
                **_json_opts,
            )
        )

    def print_mutation_result(
        self,
        item: Mapping[str, Any],
        item_name: str | None = None,
        action_name: str | None = None,
        extra_info: Mapping[str, Any] = {},
    ) -> None:
        data = {
            "ok": item.get("ok", False),
            "msg": item.get("msg", "Failed"),
            **extra_info,
        }
        if item_name is not None and item_name in item:
            data = {
                **data,
                item_name: {k: v for k, v in item[item_name].items()},
            }
        print(
            json.dumps(
                data,
                **_json_opts,
            )
        )

    def print_mutation_error(
        self,
        error: Exception | None = None,
        msg: str = "Failed",
        item_name: str | None = None,
        action_name: str | None = None,
        extra_info: Mapping[str, Any] = {},
    ) -> None:
        data = {
            "ok": False,
            "msg": msg,
            "item_name": item_name,
            "action_name": action_name,
            **extra_info,
        }
        if error is not None:
            data["error"] = str(error)
        print(
            json.dumps(
                data,
                **_json_opts,
            )
        )

    def print_error(
        self,
        error: Exception,
    ) -> None:
        match error:
            case BackendAPIError():
                print(
                    json.dumps(
                        {
                            "error": error.data["title"],
                            "api": {
                                "status": error.status,
                                "reason": error.reason,
                                **error.data,
                            },
                        },
                        **_json_opts,
                    )
                )
            case _:
                print(
                    json.dumps(
                        {
                            "error": str(error),
                        },
                        **_json_opts,
                    )
                )

    def print_fail(
        self,
        message: str,
    ) -> None:
        print(
            json.dumps(
                {
                    "error": message,
                },
                **_json_opts,
            )
        )



from __future__ import annotations

import functools
import inspect
from collections.abc import Callable, Iterable
from typing import Any

from ai.backend.client.output.types import FieldSet, FieldSpec
from ai.backend.client.session import AsyncSession, Session, api_session

__all__ = (
    "APIFunctionMeta",
    "BaseFunction",
    "api_function",
    "resolve_fields",
)


def _wrap_method(cls: type, orig_name: str, meth: Callable[..., Any]) -> Callable[..., Any]:
    @functools.wraps(meth)
    def _method(*args: Any, **kwargs: Any) -> Any:
        # We need to keep the original attributes so that they could be correctly
        # bound to the class/instance at runtime.
        func = getattr(cls, orig_name)
        coro = func(*args, **kwargs)
        _api_session = api_session.get()
        if _api_session is None:
            raise RuntimeError(
                "API functions must be called inside the context of a valid API session",
            )
        if isinstance(_api_session, AsyncSession):
            return coro
        # At this point, _api_session must be a Session (the sync version)
        if not isinstance(_api_session, Session):
            raise RuntimeError("API session must be either AsyncSession or Session")
        if inspect.isasyncgen(coro):
            return _api_session.worker_thread.execute_generator(coro)
        return _api_session.worker_thread.execute(coro)

    return _method


def api_function[T: Callable[..., Any]](meth: T) -> T:
    """
    Mark the wrapped method as the API function method.
    """
    meth._backend_api = True  # type: ignore[attr-defined]
    return meth


def resolve_fields(
    fields: Iterable[FieldSpec | str] | None,
    base_field_set: FieldSet,
    default_fields: Iterable[FieldSpec],
) -> tuple[str, ...]:
    if fields is None:
        fields = default_fields
    return tuple(
        f.field_ref if isinstance(f, FieldSpec) else base_field_set[f].field_ref for f in fields
    )


def field_resolver(
    base_field_set: FieldSet,
    default_fields: Iterable[FieldSpec],
) -> Callable[..., Any]:
    def decorator(meth: Callable[..., Any]) -> Callable[..., Any]:
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            if fields := kwargs.get("fields", default_fields):
                resolved_fields = tuple(
                    f.field_ref if isinstance(f, FieldSpec) else base_field_set[f].field_ref
                    for f in fields
                )
                kwargs["fields"] = resolved_fields
            return meth(*args, **kwargs)

        return wrapper

    return decorator


class APIFunctionMeta(type):
    """
    Converts all methods marked with :func:`api_function` into
    session-aware methods that are either plain Python functions
    or coroutines.
    """

    _async = True

    def __init__(
        cls, name: str, bases: tuple[type, ...], attrs: dict[str, Any], **kwargs: Any
    ) -> None:
        super().__init__(name, bases, attrs)
        for attr_name, attr_value in attrs.items():
            if hasattr(attr_value, "_backend_api"):
                orig_name = "_orig_" + attr_name
                setattr(cls, orig_name, attr_value)
                wrapped = _wrap_method(cls, orig_name, attr_value)
                setattr(cls, attr_name, wrapped)


class BaseFunction(metaclass=APIFunctionMeta):
    pass



from collections.abc import Mapping
from typing import Any, cast

from ai.backend.client.request import Request

from .base import BaseFunction, api_function

__all__ = ("System",)


class System(BaseFunction):
    """
    Provides the function interface for the API endpoint's system information.
    """

    @api_function
    @classmethod
    async def get_versions(cls) -> Mapping[str, str]:
        rqst = Request("GET", "/")
        async with rqst.fetch() as resp:
            return cast(Mapping[str, str], await resp.json())

    @api_function
    @classmethod
    async def get_manager_version(cls) -> str:
        rqst = Request("GET", "/")
        async with rqst.fetch() as resp:
            ret: dict[str, Any] = await resp.json()
            result: str = ret["manager"]
            return result

    @api_function
    @classmethod
    async def get_api_version(cls) -> str:
        rqst = Request("GET", "/")
        async with rqst.fetch() as resp:
            ret: dict[str, Any] = await resp.json()
            result: str = ret["version"]
            return result



"""Client SDK functions for fair share system."""

from __future__ import annotations

from decimal import Decimal
from uuid import UUID

from ai.backend.client.request import Request
from ai.backend.common.dto.manager.fair_share import (
    BulkUpsertDomainFairShareWeightRequest,
    BulkUpsertDomainFairShareWeightResponse,
    BulkUpsertProjectFairShareWeightRequest,
    BulkUpsertProjectFairShareWeightResponse,
    BulkUpsertUserFairShareWeightRequest,
    BulkUpsertUserFairShareWeightResponse,
    GetDomainFairShareResponse,
    GetProjectFairShareResponse,
    GetResourceGroupFairShareSpecResponse,
    GetUserFairShareResponse,
    ResourceWeightEntryInput,
    SearchDomainFairSharesRequest,
    SearchDomainFairSharesResponse,
    SearchProjectFairSharesRequest,
    SearchProjectFairSharesResponse,
    SearchResourceGroupFairShareSpecsResponse,
    SearchUserFairSharesRequest,
    SearchUserFairSharesResponse,
    UpdateResourceGroupFairShareSpecResponse,
    UpsertDomainFairShareWeightResponse,
    UpsertProjectFairShareWeightResponse,
    UpsertUserFairShareWeightResponse,
)

from .base import BaseFunction, api_function

__all__ = ("FairShare",)


class FairShare(BaseFunction):
    """
    Provides functions to interact with fair share data.
    Supports domain, project, and user fair shares.
    Requires superadmin privileges.
    """

    # Domain Fair Share

    @api_function
    @classmethod
    async def get_domain_fair_share(
        cls,
        resource_group: str,
        domain_name: str,
    ) -> GetDomainFairShareResponse:
        """
        Get a single domain fair share.

        :param resource_group: Resource group name
        :param domain_name: Domain name
        :returns: Domain fair share data
        """
        rqst = Request("GET", f"/fair-share/domains/{resource_group}/{domain_name}")
        async with rqst.fetch() as resp:
            data = await resp.json()
            return GetDomainFairShareResponse.model_validate(data)

    @api_function
    @classmethod
    async def search_domain_fair_shares(
        cls,
        request: SearchDomainFairSharesRequest,
    ) -> SearchDomainFairSharesResponse:
        """
        Search domain fair shares.

        :param request: Domain fair share search request
        :returns: List of domain fair shares
        """
        rqst = Request("POST", "/fair-share/domains/search")
        rqst.set_json(request.model_dump(mode="json", exclude_none=True))
        async with rqst.fetch() as resp:
            data = await resp.json()
            return SearchDomainFairSharesResponse.model_validate(data)

    # Project Fair Share

    @api_function
    @classmethod
    async def get_project_fair_share(
        cls,
        resource_group: str,
        project_id: UUID,
    ) -> GetProjectFairShareResponse:
        """
        Get a single project fair share.

        :param resource_group: Resource group name
        :param project_id: Project ID
        :returns: Project fair share data
        """
        rqst = Request("GET", f"/fair-share/projects/{resource_group}/{project_id}")
        async with rqst.fetch() as resp:
            data = await resp.json()
            return GetProjectFairShareResponse.model_validate(data)

    @api_function
    @classmethod
    async def search_project_fair_shares(
        cls,
        request: SearchProjectFairSharesRequest,
    ) -> SearchProjectFairSharesResponse:
        """
        Search project fair shares.

        :param request: Project fair share search request
        :returns: List of project fair shares
        """
        rqst = Request("POST", "/fair-share/projects/search")
        rqst.set_json(request.model_dump(mode="json", exclude_none=True))
        async with rqst.fetch() as resp:
            data = await resp.json()
            return SearchProjectFairSharesResponse.model_validate(data)

    # User Fair Share

    @api_function
    @classmethod
    async def get_user_fair_share(
        cls,
        resource_group: str,
        project_id: UUID,
        user_uuid: UUID,
    ) -> GetUserFairShareResponse:
        """
        Get a single user fair share.

        :param resource_group: Resource group name
        :param project_id: Project ID
        :param user_uuid: User UUID
        :returns: User fair share data
        """
        rqst = Request("GET", f"/fair-share/users/{resource_group}/{project_id}/{user_uuid}")
        async with rqst.fetch() as resp:
            data = await resp.json()
            return GetUserFairShareResponse.model_validate(data)

    @api_function
    @classmethod
    async def search_user_fair_shares(
        cls,
        request: SearchUserFairSharesRequest,
    ) -> SearchUserFairSharesResponse:
        """
        Search user fair shares.

        :param request: User fair share search request
        :returns: List of user fair shares
        """
        rqst = Request("POST", "/fair-share/users/search")
        rqst.set_json(request.model_dump(mode="json", exclude_none=True))
        async with rqst.fetch() as resp:
            data = await resp.json()
            return SearchUserFairSharesResponse.model_validate(data)

    # RG-Scoped Domain Fair Share (Resource Group Scope)

    @api_function
    @classmethod
    async def rg_get_domain_fair_share(
        cls,
        resource_group: str,
        domain_name: str,
    ) -> GetDomainFairShareResponse:
        """
        Get a single domain fair share within resource group scope.

        :param resource_group: Resource group name
        :param domain_name: Domain name
        :returns: Domain fair share data
        """
        rqst = Request("GET", f"/fair-share/rg/{resource_group}/domains/{domain_name}")
        async with rqst.fetch() as resp:
            data = await resp.json()
            return GetDomainFairShareResponse.model_validate(data)

    @api_function
    @classmethod
    async def rg_search_domain_fair_shares(
        cls,
        resource_group: str,
        request: SearchDomainFairSharesRequest,
    ) -> SearchDomainFairSharesResponse:
        """
        Search domain fair shares within resource group scope.

        :param resource_group: Resource group name
        :param request: Domain fair share search request
        :returns: List of domain fair shares
        """
        rqst = Request("POST", f"/fair-share/rg/{resource_group}/domains/search")
        rqst.set_json(request.model_dump(mode="json", exclude_none=True))
        async with rqst.fetch() as resp:
            data = await resp.json()
            return SearchDomainFairSharesResponse.model_validate(data)

    # RG-Scoped Project Fair Share (Resource Group Scope)

    @api_function
    @classmethod
    async def rg_get_project_fair_share(
        cls,
        resource_group: str,
        domain_name: str,
        project_id: UUID,
    ) -> GetProjectFairShareResponse:
        """
        Get a single project fair share within resource group scope.

        :param resource_group: Resource group name
        :param domain_name: Domain name
        :param project_id: Project ID
        :returns: Project fair share data
        """
        rqst = Request(
            "GET", f"/fair-share/rg/{resource_group}/domains/{domain_name}/projects/{project_id}"
        )
        async with rqst.fetch() as resp:
            data = await resp.json()
            return GetProjectFairShareResponse.model_validate(data)

    @api_function
    @classmethod
    async def rg_search_project_fair_shares(
        cls,
        resource_group: str,
        domain_name: str,
        request: SearchProjectFairSharesRequest,
    ) -> SearchProjectFairSharesResponse:
        """
        Search project fair shares within resource group scope.

        :param resource_group: Resource group name
        :param domain_name: Domain name
        :param request: Project fair share search request
        :returns: List of project fair shares
        """
        rqst = Request(
            "POST", f"/fair-share/rg/{resource_group}/domains/{domain_name}/projects/search"
        )
        rqst.set_json(request.model_dump(mode="json", exclude_none=True))
        async with rqst.fetch() as resp:
            data = await resp.json()
            return SearchProjectFairSharesResponse.model_validate(data)

    # RG-Scoped User Fair Share (Resource Group Scope)

    @api_function
    @classmethod
    async def rg_get_user_fair_share(
        cls,
        resource_group: str,
        domain_name: str,
        project_id: UUID,
        user_uuid: UUID,
    ) -> GetUserFairShareResponse:
        """
        Get a single user fair share within resource group scope.

        :param resource_group: Resource group name
        :param domain_name: Domain name
        :param project_id: Project ID
        :param user_uuid: User UUID
        :returns: User fair share data
        """
        rqst = Request(
            "GET",
            f"/fair-share/rg/{resource_group}/domains/{domain_name}/projects/{project_id}/users/{user_uuid}",
        )
        async with rqst.fetch() as resp:
            data = await resp.json()
            return GetUserFairShareResponse.model_validate(data)

    @api_function
    @classmethod
    async def rg_search_user_fair_shares(
        cls,
        resource_group: str,
        domain_name: str,
        project_id: UUID,
        request: SearchUserFairSharesRequest,
    ) -> SearchUserFairSharesResponse:
        """
        Search user fair shares within resource group scope.

        :param resource_group: Resource group name
        :param domain_name: Domain name
        :param project_id: Project ID
        :param request: User fair share search request
        :returns: List of user fair shares
        """
        rqst = Request(
            "POST",
            f"/fair-share/rg/{resource_group}/domains/{domain_name}/projects/{project_id}/users/search",
        )
        rqst.set_json(request.model_dump(mode="json", exclude_none=True))
        async with rqst.fetch() as resp:
            data = await resp.json()
            return SearchUserFairSharesResponse.model_validate(data)

    # Upsert Weight Methods

    @api_function
    @classmethod
    async def upsert_domain_fair_share_weight(
        cls,
        resource_group: str,
        domain_name: str,
        weight: Decimal | None = None,
    ) -> UpsertDomainFairShareWeightResponse:
        """
        Upsert domain fair share weight.

        Creates a new domain fair share record or updates the weight of an existing one.
        If weight is None, the resource group's default_weight will be used.

        :param resource_group: Resource group (scaling group) name
        :param domain_name: Domain name
        :param weight: Weight value (None to use resource group's default_weight)
        :returns: Updated domain fair share data
        """
        rqst = Request("PUT", f"/fair-share/domains/{resource_group}/{domain_name}/weight")
        body: dict[str, str | None] = {"weight": str(weight) if weight is not None else None}
        rqst.set_json(body)
        async with rqst.fetch() as resp:
            data = await resp.json()
            return UpsertDomainFairShareWeightResponse.model_validate(data)

    @api_function
    @classmethod
    async def upsert_project_fair_share_weight(
        cls,
        resource_group: str,
        project_id: UUID,
        domain_name: str,
        weight: Decimal | None = None,
    ) -> UpsertProjectFairShareWeightResponse:
        """
        Upsert project fair share weight.

        Creates a new project fair share record or updates the weight of an existing one.
        If weight is None, the resource group's default_weight will be used.

        :param resource_group: Resource group (scaling group) name
        :param project_id: Project ID
        :param domain_name: Domain name the project belongs to
        :param weight: Weight value (None to use resource group's default_weight)
        :returns: Updated project fair share data
        """
        rqst = Request("PUT", f"/fair-share/projects/{resource_group}/{project_id}/weight")
        body: dict[str, str | None] = {
            "domain_name": domain_name,
            "weight": str(weight) if weight is not None else None,
        }
        rqst.set_json(body)
        async with rqst.fetch() as resp:
            data = await resp.json()
            return UpsertProjectFairShareWeightResponse.model_validate(data)

    @api_function
    @classmethod
    async def upsert_user_fair_share_weight(
        cls,
        resource_group: str,
        project_id: UUID,
        user_uuid: UUID,
        domain_name: str,
        weight: Decimal | None = None,
    ) -> UpsertUserFairShareWeightResponse:
        """
        Upsert user fair share weight.

        Creates a new user fair share record or updates the weight of an existing one.
        If weight is None, the resource group's default_weight will be used.

        :param resource_group: Resource group (scaling group) name
        :param project_id: Project ID
        :param user_uuid: User UUID
        :param domain_name: Domain name the user belongs to
        :param weight: Weight value (None to use resource group's default_weight)
        :returns: Updated user fair share data
        """
        rqst = Request("PUT", f"/fair-share/users/{resource_group}/{project_id}/{user_uuid}/weight")
        body: dict[str, str | None] = {
            "domain_name": domain_name,
            "weight": str(weight) if weight is not None else None,
        }
        rqst.set_json(body)
        async with rqst.fetch() as resp:
            data = await resp.json()
            return UpsertUserFairShareWeightResponse.model_validate(data)

    # Bulk Upsert Weight Methods

    @api_function
    @classmethod
    async def bulk_upsert_domain_fair_share_weight(
        cls,
        request: BulkUpsertDomainFairShareWeightRequest,
    ) -> BulkUpsertDomainFairShareWeightResponse:
        """
        Bulk upsert domain fair share weights.

        Creates new domain fair share records or updates weights of existing ones.
        If weight is None for an entry, the resource group's default_weight will be used.

        :param request: Bulk upsert domain weight request
        :returns: Bulk upsert result with count
        """
        rqst = Request("POST", "/fair-share/domains/bulk-upsert-weight")
        rqst.set_json(request.model_dump(mode="json", exclude_none=True))
        async with rqst.fetch() as resp:
            data = await resp.json()
            return BulkUpsertDomainFairShareWeightResponse.model_validate(data)

    # Resource Group Fair Share Spec Methods

    @api_function
    @classmethod
    async def bulk_upsert_project_fair_share_weight(
        cls,
        request: BulkUpsertProjectFairShareWeightRequest,
    ) -> BulkUpsertProjectFairShareWeightResponse:
        """
        Bulk upsert project fair share weights.

        Creates new project fair share records or updates weights of existing ones.
        If weight is None for an entry, the resource group's default_weight will be used.

        :param request: Bulk upsert project weight request
        :returns: Bulk upsert result with count
        """
        rqst = Request("POST", "/fair-share/projects/bulk-upsert-weight")
        rqst.set_json(request.model_dump(mode="json", exclude_none=True))
        async with rqst.fetch() as resp:
            data = await resp.json()
            return BulkUpsertProjectFairShareWeightResponse.model_validate(data)

    @api_function
    @classmethod
    async def bulk_upsert_user_fair_share_weight(
        cls,
        request: BulkUpsertUserFairShareWeightRequest,
    ) -> BulkUpsertUserFairShareWeightResponse:
        """
        Bulk upsert user fair share weights.

        Creates new user fair share records or updates weights of existing ones.
        If weight is None for an entry, the resource group's default_weight will be used.

        :param request: Bulk upsert user weight request
        :returns: Bulk upsert result with count
        """
        rqst = Request("POST", "/fair-share/users/bulk-upsert-weight")
        rqst.set_json(request.model_dump(mode="json", exclude_none=True))
        async with rqst.fetch() as resp:
            data = await resp.json()
            return BulkUpsertUserFairShareWeightResponse.model_validate(data)

    # Resource Group Fair Share Spec Methods

    @api_function
    @classmethod
    async def update_resource_group_fair_share_spec(
        cls,
        resource_group: str,
        half_life_days: int | None = None,
        lookback_days: int | None = None,
        decay_unit_days: int | None = None,
        default_weight: Decimal | None = None,
        resource_weights: list[ResourceWeightEntryInput] | None = None,
    ) -> UpdateResourceGroupFairShareSpecResponse:
        """
        Update resource group fair share specification.

        Performs a partial update - only provided fields are updated.
        Other fields retain their existing values.

        :param resource_group: Resource group (scaling group) name
        :param half_life_days: Half-life for exponential decay in days
        :param lookback_days: Total lookback period in days
        :param decay_unit_days: Granularity of decay buckets in days
        :param default_weight: Default weight for entities
        :param resource_weights: List of resource weight entries (resource_type and weight)
        :returns: Updated resource group fair share spec
        """
        rqst = Request("PATCH", f"/fair-share/resource-groups/{resource_group}/spec")
        body: dict[str, int | str | list[dict[str, str | None]] | None] = {}
        if half_life_days is not None:
            body["half_life_days"] = half_life_days
        if lookback_days is not None:
            body["lookback_days"] = lookback_days
        if decay_unit_days is not None:
            body["decay_unit_days"] = decay_unit_days
        if default_weight is not None:
            body["default_weight"] = str(default_weight)
        if resource_weights is not None:
            body["resource_weights"] = [
                {
                    "resource_type": entry.resource_type,
                    "weight": str(entry.weight) if entry.weight is not None else None,
                }
                for entry in resource_weights
            ]
        rqst.set_json(body)
        async with rqst.fetch() as resp:
            data = await resp.json()
            return UpdateResourceGroupFairShareSpecResponse.model_validate(data)

    @api_function
    @classmethod
    async def get_resource_group_fair_share_spec(
        cls,
        resource_group: str,
    ) -> GetResourceGroupFairShareSpecResponse:
        """
        Get resource group fair share specification.

        :param resource_group: Resource group (scaling group) name
        :returns: Resource group fair share spec
        """
        rqst = Request("GET", f"/fair-share/resource-groups/{resource_group}/spec")
        async with rqst.fetch() as resp:
            data = await resp.json()
            return GetResourceGroupFairShareSpecResponse.model_validate(data)

    @api_function
    @classmethod
    async def list_resource_group_fair_share_specs(
        cls,
    ) -> SearchResourceGroupFairShareSpecsResponse:
        """
        List all resource groups with their fair share specifications.

        :returns: List of resource groups with fair share specs
        """
        rqst = Request("GET", "/fair-share/resource-groups/specs")
        async with rqst.fetch() as resp:
            data = await resp.json()
            return SearchResourceGroupFairShareSpecsResponse.model_validate(data)



"""Client SDK functions for CSV export system."""

from __future__ import annotations

from collections.abc import AsyncIterator

from ai.backend.client.request import Request
from ai.backend.common.dto.manager.export import (
    AuditLogExportCSVRequest,
    AuditLogExportFilter,
    AuditLogExportOrder,
    GetExportReportResponse,
    KeypairExportCSVRequest,
    ListExportReportsResponse,
    ProjectExportCSVRequest,
    ProjectExportFilter,
    ProjectExportOrder,
    SessionExportCSVRequest,
    SessionExportFilter,
    SessionExportOrder,
    UserExportCSVRequest,
    UserExportFilter,
    UserExportOrder,
)

from .base import BaseFunction, api_function

__all__ = ("Export",)


class Export(BaseFunction):
    """
    Provides functions to interact with the CSV export system.
    Supports listing available reports and streaming report-specific CSV exports.
    """

    @api_function
    @classmethod
    async def list_reports(cls) -> ListExportReportsResponse:
        """
        List all available export reports.

        :returns: List of available export reports with their fields
        """
        rqst = Request("GET", "/export/reports")
        async with rqst.fetch() as resp:
            data = await resp.json()
            return ListExportReportsResponse.model_validate(data)

    @api_function
    @classmethod
    async def get_report(cls, report_key: str) -> GetExportReportResponse:
        """
        Get a specific export report by key.

        :param report_key: The report key (e.g., 'sessions', 'users', 'projects')
        :returns: Report details including available fields
        """
        rqst = Request("GET", f"/export/reports/{report_key}")
        async with rqst.fetch() as resp:
            data = await resp.json()
            return GetExportReportResponse.model_validate(data)

    # =========================================================================
    # User Export
    # =========================================================================

    @api_function
    @classmethod
    async def stream_users_csv(
        cls,
        *,
        fields: list[str] | None = None,
        filter: UserExportFilter | None = None,
        order: list[UserExportOrder] | None = None,
        encoding: str = "utf-8",
        filename: str | None = None,
        chunk_size: int = 8192,
    ) -> AsyncIterator[bytes]:
        """
        Stream user export as an async iterator of chunks.

        :param fields: Optional list of field keys to include (default: all fields)
        :param filter: Optional user-specific filter conditions
        :param order: Optional list of user order specifications
        :param encoding: CSV encoding (default: utf-8, also supports euc-kr)
        :param filename: Optional filename for the export
        :param chunk_size: Size of chunks to yield (default: 8192 bytes)
        :yields: Chunks of CSV data as bytes
        """
        request = UserExportCSVRequest(
            fields=fields,
            filter=filter,
            order=order,
            encoding=encoding,
        )

        rqst = Request("POST", "/export/users/csv")
        rqst.set_json(request.model_dump(mode="json", exclude_none=True))
        if filename:
            rqst.headers["X-Export-Filename"] = filename

        async with rqst.fetch() as resp:
            async for chunk in resp.content.iter_chunked(chunk_size):
                yield chunk

    # =========================================================================
    # Session Export
    # =========================================================================

    @api_function
    @classmethod
    async def stream_sessions_csv(
        cls,
        *,
        fields: list[str] | None = None,
        filter: SessionExportFilter | None = None,
        order: list[SessionExportOrder] | None = None,
        encoding: str = "utf-8",
        filename: str | None = None,
        chunk_size: int = 8192,
    ) -> AsyncIterator[bytes]:
        """
        Stream session export as an async iterator of chunks.

        :param fields: Optional list of field keys to include (default: all fields)
        :param filter: Optional session-specific filter conditions
        :param order: Optional list of session order specifications
        :param encoding: CSV encoding (default: utf-8, also supports euc-kr)
        :param filename: Optional filename for the export
        :param chunk_size: Size of chunks to yield (default: 8192 bytes)
        :yields: Chunks of CSV data as bytes
        """
        request = SessionExportCSVRequest(
            fields=fields,
            filter=filter,
            order=order,
            encoding=encoding,
        )

        rqst = Request("POST", "/export/sessions/csv")
        rqst.set_json(request.model_dump(mode="json", exclude_none=True))
        if filename:
            rqst.headers["X-Export-Filename"] = filename

        async with rqst.fetch() as resp:
            async for chunk in resp.content.iter_chunked(chunk_size):
                yield chunk

    # =========================================================================
    # Project Export
    # =========================================================================

    @api_function
    @classmethod
    async def stream_projects_csv(
        cls,
        *,
        fields: list[str] | None = None,
        filter: ProjectExportFilter | None = None,
        order: list[ProjectExportOrder] | None = None,
        encoding: str = "utf-8",
        filename: str | None = None,
        chunk_size: int = 8192,
    ) -> AsyncIterator[bytes]:
        """
        Stream project export as an async iterator of chunks.

        :param fields: Optional list of field keys to include (default: all fields)
        :param filter: Optional project-specific filter conditions
        :param order: Optional list of project order specifications
        :param encoding: CSV encoding (default: utf-8, also supports euc-kr)
        :param filename: Optional filename for the export
        :param chunk_size: Size of chunks to yield (default: 8192 bytes)
        :yields: Chunks of CSV data as bytes
        """
        request = ProjectExportCSVRequest(
            fields=fields,
            filter=filter,
            order=order,
            encoding=encoding,
        )

        rqst = Request("POST", "/export/projects/csv")
        rqst.set_json(request.model_dump(mode="json", exclude_none=True))
        if filename:
            rqst.headers["X-Export-Filename"] = filename

        async with rqst.fetch() as resp:
            async for chunk in resp.content.iter_chunked(chunk_size):
                yield chunk

    # =========================================================================
    # Keypair Export
    # =========================================================================

    @api_function
    @classmethod
    async def stream_keypairs_csv(
        cls,
        *,
        fields: list[str] | None = None,
        encoding: str = "utf-8",
        filename: str | None = None,
        chunk_size: int = 8192,
    ) -> AsyncIterator[bytes]:
        """
        Stream keypair export as an async iterator of chunks.

        Note: Keypair export does not currently support filtering or ordering.

        :param fields: Optional list of field keys to include (default: all fields)
        :param encoding: CSV encoding (default: utf-8, also supports euc-kr)
        :param filename: Optional filename for the export
        :param chunk_size: Size of chunks to yield (default: 8192 bytes)
        :yields: Chunks of CSV data as bytes
        """
        request = KeypairExportCSVRequest(
            fields=fields,
            encoding=encoding,
        )

        rqst = Request("POST", "/export/keypairs/csv")
        rqst.set_json(request.model_dump(mode="json", exclude_none=True))
        if filename:
            rqst.headers["X-Export-Filename"] = filename

        async with rqst.fetch() as resp:
            async for chunk in resp.content.iter_chunked(chunk_size):
                yield chunk

    # =========================================================================
    # Audit Log Export
    # =========================================================================

    @api_function
    @classmethod
    async def stream_audit_logs_csv(
        cls,
        *,
        fields: list[str] | None = None,
        filter: AuditLogExportFilter | None = None,
        order: list[AuditLogExportOrder] | None = None,
        encoding: str = "utf-8",
        filename: str | None = None,
        chunk_size: int = 8192,
    ) -> AsyncIterator[bytes]:
        """
        Stream audit log export as an async iterator of chunks.

        :param fields: Optional list of field keys to include (default: all fields)
        :param filter: Optional audit log-specific filter conditions
        :param order: Optional list of audit log order specifications
        :param encoding: CSV encoding (default: utf-8, also supports euc-kr)
        :param filename: Optional filename for the export
        :param chunk_size: Size of chunks to yield (default: 8192 bytes)
        :yields: Chunks of CSV data as bytes
        """
        request = AuditLogExportCSVRequest(
            fields=fields,
            filter=filter,
            order=order,
            encoding=encoding,
        )

        rqst = Request("POST", "/export/audit-logs/csv")
        rqst.set_json(request.model_dump(mode="json", exclude_none=True))
        if filename:
            rqst.headers["X-Export-Filename"] = filename

        async with rqst.fetch() as resp:
            async for chunk in resp.content.iter_chunked(chunk_size):
                yield chunk



"""CLI commands for fair share scheduler."""

from __future__ import annotations

import json
import sys
from decimal import Decimal
from uuid import UUID

import click

from ai.backend.client.cli.extensions import pass_ctx_obj
from ai.backend.client.cli.fair_share import fair_share
from ai.backend.client.cli.types import CLIContext

# =============================================================================
# Domain Fair Share Commands
# =============================================================================


@fair_share.group()
def domain() -> None:
    """Domain-level fair share operations."""


@domain.command("get")
@pass_ctx_obj
@click.argument("resource_group", type=str)
@click.argument("domain_name", type=str)
@click.option("--json", "as_json", is_flag=True, help="Output as JSON")
def domain_get_cmd(
    ctx: CLIContext,
    resource_group: str,
    domain_name: str,
    as_json: bool,
) -> None:
    """Get domain fair share data."""
    from ai.backend.cli.types import ExitCode
    from ai.backend.client.session import Session

    with Session() as api_session:
        try:
            response = api_session.FairShare.get_domain_fair_share(resource_group, domain_name)
            if response.item is None:
                print("No domain fair share found")
                return

            if as_json:
                print(json.dumps(response.item.model_dump(mode="json"), indent=2, default=str))
            else:
                fs = response.item
                print(f"Resource Group: {fs.resource_group}")
                print(f"Domain: {fs.domain_name}")
                print(f"Weight: {fs.spec.weight}")
                print(f"Fair Share Factor: {fs.calculation_snapshot.fair_share_factor}")
                print(f"Normalized Usage: {fs.calculation_snapshot.normalized_usage}")
                print(
                    f"Lookback: {fs.calculation_snapshot.lookback_start} ~ "
                    f"{fs.calculation_snapshot.lookback_end}"
                )
                print(f"Created: {fs.created_at}")
                print(f"Updated: {fs.updated_at}")
        except Exception as e:
            ctx.output.print_error(e)
            sys.exit(ExitCode.FAILURE)


@domain.command("list")
@pass_ctx_obj
@click.option("--resource-group", type=str, default=None, help="Filter by resource group")
@click.option("--domain-name", type=str, default=None, help="Filter by domain name")
@click.option("--limit", type=int, default=20, help="Maximum number of records to return")
@click.option("--offset", type=int, default=0, help="Offset for pagination")
@click.option(
    "--order-by",
    type=click.Choice(["fair_share_factor", "domain_name", "created_at"]),
    default="fair_share_factor",
    help="Order by field",
)
@click.option(
    "--order",
    type=click.Choice(["ASC", "DESC"]),
    default="DESC",
    help="Order direction",
)
@click.option("--json", "as_json", is_flag=True, help="Output as JSON")
def domain_list_cmd(
    ctx: CLIContext,
    resource_group: str | None,
    domain_name: str | None,
    limit: int,
    offset: int,
    order_by: str,
    order: str,
    as_json: bool,
) -> None:
    """List domain fair shares."""
    from ai.backend.cli.types import ExitCode
    from ai.backend.client.session import Session
    from ai.backend.common.dto.manager.fair_share import (
        DomainFairShareFilter,
        DomainFairShareOrder,
        DomainFairShareOrderField,
        OrderDirection,
        SearchDomainFairSharesRequest,
    )
    from ai.backend.common.dto.manager.query import StringFilter

    with Session() as api_session:
        try:
            filter_cond = None
            if resource_group or domain_name:
                filter_cond = DomainFairShareFilter(
                    resource_group=StringFilter(equals=resource_group) if resource_group else None,
                    domain_name=StringFilter(equals=domain_name) if domain_name else None,
                )

            order_spec = DomainFairShareOrder(
                field=DomainFairShareOrderField(order_by),
                direction=OrderDirection(order.lower()),
            )

            request = SearchDomainFairSharesRequest(
                filter=filter_cond,
                order=[order_spec],
                limit=limit,
                offset=offset,
            )
            response = api_session.FairShare.search_domain_fair_shares(request)

            if as_json:
                print(
                    json.dumps(
                        [fs.model_dump(mode="json") for fs in response.items],
                        indent=2,
                        default=str,
                    )
                )
            else:
                items = response.items
                if not items:
                    print("No domain fair shares found")
                    return
                print(f"Total: {response.pagination.total}")
                print()
                for fs in items:
                    print(f"ID: {fs.id}")
                    print(f"Resource Group: {fs.resource_group}")
                    print(f"Domain: {fs.domain_name}")
                    print(f"Weight: {fs.spec.weight}")
                    print(f"Fair Share Factor: {fs.calculation_snapshot.fair_share_factor}")
                    print(f"Normalized Usage: {fs.calculation_snapshot.normalized_usage}")
                    print(f"Created: {fs.created_at}")
                    print("---")
        except Exception as e:
            ctx.output.print_error(e)
            sys.exit(ExitCode.FAILURE)


@domain.command("set-weight")
@pass_ctx_obj
@click.argument("resource_group", type=str)
@click.argument("domain_name", type=str)
@click.option(
    "--weight",
    type=str,
    default=None,
    help="Weight value (omit to use resource group's default_weight)",
)
@click.option("--json", "as_json", is_flag=True, help="Output as JSON")
def domain_set_weight_cmd(
    ctx: CLIContext,
    resource_group: str,
    domain_name: str,
    weight: str | None,
    as_json: bool,
) -> None:
    """Set domain fair share weight."""
    from ai.backend.cli.types import ExitCode
    from ai.backend.client.session import Session

    with Session() as api_session:
        try:
            weight_decimal = Decimal(weight) if weight is not None else None
            response = api_session.FairShare.upsert_domain_fair_share_weight(
                resource_group, domain_name, weight_decimal
            )

            if as_json:
                print(json.dumps(response.item.model_dump(mode="json"), indent=2, default=str))
            else:
                fs = response.item
                print("Updated domain fair share weight successfully.")
                print(f"ID: {fs.id}")
                print(f"Resource Group: {fs.resource_group}")
                print(f"Domain: {fs.domain_name}")
                print(f"Weight: {fs.spec.weight}")
                print(f"Updated: {fs.updated_at}")
        except Exception as e:
            ctx.output.print_error(e)
            sys.exit(ExitCode.FAILURE)


# =============================================================================
# Project Fair Share Commands
# =============================================================================


@fair_share.group()
def project() -> None:
    """Project-level fair share operations."""


@project.command("get")
@pass_ctx_obj
@click.argument("resource_group", type=str)
@click.argument("project_id", type=str)
@click.option("--json", "as_json", is_flag=True, help="Output as JSON")
def project_get_cmd(
    ctx: CLIContext,
    resource_group: str,
    project_id: str,
    as_json: bool,
) -> None:
    """Get project fair share data."""
    from ai.backend.cli.types import ExitCode
    from ai.backend.client.session import Session

    with Session() as api_session:
        try:
            response = api_session.FairShare.get_project_fair_share(
                resource_group, UUID(project_id)
            )
            if response.item is None:
                print("No project fair share found")
                return

            if as_json:
                print(json.dumps(response.item.model_dump(mode="json"), indent=2, default=str))
            else:
                fs = response.item
                print(f"Resource Group: {fs.resource_group}")
                print(f"Project ID: {fs.project_id}")
                print(f"Domain: {fs.domain_name}")
                print(f"Weight: {fs.spec.weight}")
                print(f"Fair Share Factor: {fs.calculation_snapshot.fair_share_factor}")
                print(f"Normalized Usage: {fs.calculation_snapshot.normalized_usage}")
                print(
                    f"Lookback: {fs.calculation_snapshot.lookback_start} ~ "
                    f"{fs.calculation_snapshot.lookback_end}"
                )
                print(f"Created: {fs.created_at}")
                print(f"Updated: {fs.updated_at}")
        except Exception as e:
            ctx.output.print_error(e)
            sys.exit(ExitCode.FAILURE)


@project.command("list")
@pass_ctx_obj
@click.option("--resource-group", type=str, default=None, help="Filter by resource group")
@click.option("--project-id", type=str, default=None, help="Filter by project ID")
@click.option("--domain-name", type=str, default=None, help="Filter by domain name")
@click.option("--limit", type=int, default=20, help="Maximum number of records to return")
@click.option("--offset", type=int, default=0, help="Offset for pagination")
@click.option(
    "--order-by",
    type=click.Choice(["fair_share_factor", "created_at"]),
    default="fair_share_factor",
    help="Order by field",
)
@click.option(
    "--order",
    type=click.Choice(["ASC", "DESC"]),
    default="DESC",
    help="Order direction",
)
@click.option("--json", "as_json", is_flag=True, help="Output as JSON")
def project_list_cmd(
    ctx: CLIContext,
    resource_group: str | None,
    project_id: str | None,
    domain_name: str | None,
    limit: int,
    offset: int,
    order_by: str,
    order: str,
    as_json: bool,
) -> None:
    """List project fair shares."""
    from ai.backend.cli.types import ExitCode
    from ai.backend.client.session import Session
    from ai.backend.common.dto.manager.fair_share import (
        OrderDirection,
        ProjectFairShareFilter,
        ProjectFairShareOrder,
        ProjectFairShareOrderField,
        SearchProjectFairSharesRequest,
    )
    from ai.backend.common.dto.manager.query import StringFilter, UUIDFilter

    with Session() as api_session:
        try:
            filter_cond = None
            if resource_group or project_id or domain_name:
                filter_cond = ProjectFairShareFilter(
                    resource_group=StringFilter(equals=resource_group) if resource_group else None,
                    project_id=UUIDFilter(equals=UUID(project_id)) if project_id else None,
                    domain_name=StringFilter(equals=domain_name) if domain_name else None,
                )

            order_spec = ProjectFairShareOrder(
                field=ProjectFairShareOrderField(order_by),
                direction=OrderDirection(order.lower()),
            )

            request = SearchProjectFairSharesRequest(
                filter=filter_cond,
                order=[order_spec],
                limit=limit,
                offset=offset,
            )
            response = api_session.FairShare.search_project_fair_shares(request)

            if as_json:
                print(
                    json.dumps(
                        [fs.model_dump(mode="json") for fs in response.items],
                        indent=2,
                        default=str,
                    )
                )
            else:
                items = response.items
                if not items:
                    print("No project fair shares found")
                    return
                print(f"Total: {response.pagination.total}")
                print()
                for fs in items:
                    print(f"ID: {fs.id}")
                    print(f"Resource Group: {fs.resource_group}")
                    print(f"Project ID: {fs.project_id}")
                    print(f"Domain: {fs.domain_name}")
                    print(f"Weight: {fs.spec.weight}")
                    print(f"Fair Share Factor: {fs.calculation_snapshot.fair_share_factor}")
                    print(f"Normalized Usage: {fs.calculation_snapshot.normalized_usage}")
                    print(f"Created: {fs.created_at}")
                    print("---")
        except Exception as e:
            ctx.output.print_error(e)
            sys.exit(ExitCode.FAILURE)


@project.command("set-weight")
@pass_ctx_obj
@click.argument("resource_group", type=str)
@click.argument("project_id", type=str)
@click.option(
    "--domain",
    "domain_name",
    type=str,
    required=True,
    help="Domain name the project belongs to",
)
@click.option(
    "--weight",
    type=str,
    default=None,
    help="Weight value (omit to use resource group's default_weight)",
)
@click.option("--json", "as_json", is_flag=True, help="Output as JSON")
def project_set_weight_cmd(
    ctx: CLIContext,
    resource_group: str,
    project_id: str,
    domain_name: str,
    weight: str | None,
    as_json: bool,
) -> None:
    """Set project fair share weight."""
    from ai.backend.cli.types import ExitCode
    from ai.backend.client.session import Session

    with Session() as api_session:
        try:
            weight_decimal = Decimal(weight) if weight is not None else None
            response = api_session.FairShare.upsert_project_fair_share_weight(
                resource_group, UUID(project_id), domain_name, weight_decimal
            )

            if as_json:
                print(json.dumps(response.item.model_dump(mode="json"), indent=2, default=str))
            else:
                fs = response.item
                print("Updated project fair share weight successfully.")
                print(f"ID: {fs.id}")
                print(f"Resource Group: {fs.resource_group}")
                print(f"Project ID: {fs.project_id}")
                print(f"Domain: {fs.domain_name}")
                print(f"Weight: {fs.spec.weight}")
                print(f"Updated: {fs.updated_at}")
        except Exception as e:
            ctx.output.print_error(e)
            sys.exit(ExitCode.FAILURE)


# =============================================================================
# User Fair Share Commands
# =============================================================================


@fair_share.group()
def user() -> None:
    """User-level fair share operations."""


@user.command("get")
@pass_ctx_obj
@click.argument("resource_group", type=str)
@click.argument("project_id", type=str)
@click.argument("user_uuid", type=str)
@click.option("--json", "as_json", is_flag=True, help="Output as JSON")
def user_get_cmd(
    ctx: CLIContext,
    resource_group: str,
    project_id: str,
    user_uuid: str,
    as_json: bool,
) -> None:
    """Get user fair share data."""
    from ai.backend.cli.types import ExitCode
    from ai.backend.client.session import Session

    with Session() as api_session:
        try:
            response = api_session.FairShare.get_user_fair_share(
                resource_group, UUID(project_id), UUID(user_uuid)
            )
            if response.item is None:
                print("No user fair share found")
                return

            if as_json:
                print(json.dumps(response.item.model_dump(mode="json"), indent=2, default=str))
            else:
                fs = response.item
                print(f"Resource Group: {fs.resource_group}")
                print(f"User UUID: {fs.user_uuid}")
                print(f"Project ID: {fs.project_id}")
                print(f"Domain: {fs.domain_name}")
                print(f"Weight: {fs.spec.weight}")
                print(f"Fair Share Factor: {fs.calculation_snapshot.fair_share_factor}")
                print(f"Normalized Usage: {fs.calculation_snapshot.normalized_usage}")
                print(
                    f"Lookback: {fs.calculation_snapshot.lookback_start} ~ "
                    f"{fs.calculation_snapshot.lookback_end}"
                )
                print(f"Created: {fs.created_at}")
                print(f"Updated: {fs.updated_at}")
        except Exception as e:
            ctx.output.print_error(e)
            sys.exit(ExitCode.FAILURE)


@user.command("list")
@pass_ctx_obj
@click.option("--resource-group", type=str, default=None, help="Filter by resource group")
@click.option("--user-uuid", type=str, default=None, help="Filter by user UUID")
@click.option("--project-id", type=str, default=None, help="Filter by project ID")
@click.option("--domain-name", type=str, default=None, help="Filter by domain name")
@click.option("--limit", type=int, default=20, help="Maximum number of records to return")
@click.option("--offset", type=int, default=0, help="Offset for pagination")
@click.option(
    "--order-by",
    type=click.Choice(["fair_share_factor", "created_at"]),
    default="fair_share_factor",
    help="Order by field",
)
@click.option(
    "--order",
    type=click.Choice(["ASC", "DESC"]),
    default="DESC",
    help="Order direction",
)
@click.option("--json", "as_json", is_flag=True, help="Output as JSON")
def user_list_cmd(
    ctx: CLIContext,
    resource_group: str | None,
    user_uuid: str | None,
    project_id: str | None,
    domain_name: str | None,
    limit: int,
    offset: int,
    order_by: str,
    order: str,
    as_json: bool,
) -> None:
    """List user fair shares."""
    from ai.backend.cli.types import ExitCode
    from ai.backend.client.session import Session
    from ai.backend.common.dto.manager.fair_share import (
        OrderDirection,
        SearchUserFairSharesRequest,
        UserFairShareFilter,
        UserFairShareOrder,
        UserFairShareOrderField,
    )
    from ai.backend.common.dto.manager.query import StringFilter, UUIDFilter

    with Session() as api_session:
        try:
            filter_cond = None
            if resource_group or user_uuid or project_id or domain_name:
                filter_cond = UserFairShareFilter(
                    resource_group=StringFilter(equals=resource_group) if resource_group else None,
                    user_uuid=UUIDFilter(equals=UUID(user_uuid)) if user_uuid else None,
                    project_id=UUIDFilter(equals=UUID(project_id)) if project_id else None,
                    domain_name=StringFilter(equals=domain_name) if domain_name else None,
                )

            order_spec = UserFairShareOrder(
                field=UserFairShareOrderField(order_by),
                direction=OrderDirection(order.lower()),
            )

            request = SearchUserFairSharesRequest(
                filter=filter_cond,
                order=[order_spec],
                limit=limit,
                offset=offset,
            )
            response = api_session.FairShare.search_user_fair_shares(request)

            if as_json:
                print(
                    json.dumps(
                        [fs.model_dump(mode="json") for fs in response.items],
                        indent=2,
                        default=str,
                    )
                )
            else:
                items = response.items
                if not items:
                    print("No user fair shares found")
                    return
                print(f"Total: {response.pagination.total}")
                print()
                for fs in items:
                    print(f"ID: {fs.id}")
                    print(f"Resource Group: {fs.resource_group}")
                    print(f"User UUID: {fs.user_uuid}")
                    print(f"Project ID: {fs.project_id}")
                    print(f"Domain: {fs.domain_name}")
                    print(f"Weight: {fs.spec.weight}")
                    print(f"Fair Share Factor: {fs.calculation_snapshot.fair_share_factor}")
                    print(f"Normalized Usage: {fs.calculation_snapshot.normalized_usage}")
                    print(f"Created: {fs.created_at}")
                    print("---")
        except Exception as e:
            ctx.output.print_error(e)
            sys.exit(ExitCode.FAILURE)


@user.command("set-weight")
@pass_ctx_obj
@click.argument("resource_group", type=str)
@click.argument("project_id", type=str)
@click.argument("user_uuid", type=str)
@click.option(
    "--domain",
    "domain_name",
    type=str,
    required=True,
    help="Domain name the user belongs to",
)
@click.option(
    "--weight",
    type=str,
    default=None,
    help="Weight value (omit to use resource group's default_weight)",
)
@click.option("--json", "as_json", is_flag=True, help="Output as JSON")
def user_set_weight_cmd(
    ctx: CLIContext,
    resource_group: str,
    project_id: str,
    user_uuid: str,
    domain_name: str,
    weight: str | None,
    as_json: bool,
) -> None:
    """Set user fair share weight."""
    from ai.backend.cli.types import ExitCode
    from ai.backend.client.session import Session

    with Session() as api_session:
        try:
            weight_decimal = Decimal(weight) if weight is not None else None
            response = api_session.FairShare.upsert_user_fair_share_weight(
                resource_group, UUID(project_id), UUID(user_uuid), domain_name, weight_decimal
            )

            if as_json:
                print(json.dumps(response.item.model_dump(mode="json"), indent=2, default=str))
            else:
                fs = response.item
                print("Updated user fair share weight successfully.")
                print(f"ID: {fs.id}")
                print(f"Resource Group: {fs.resource_group}")
                print(f"User UUID: {fs.user_uuid}")
                print(f"Project ID: {fs.project_id}")
                print(f"Domain: {fs.domain_name}")
                print(f"Weight: {fs.spec.weight}")
                print(f"Updated: {fs.updated_at}")
        except Exception as e:
            ctx.output.print_error(e)
            sys.exit(ExitCode.FAILURE)


# =============================================================================
# Resource Group Fair Share Commands
# =============================================================================


@fair_share.group()
def resource_group() -> None:
    """Resource group fair share operations."""


@resource_group.command("update-spec")
@pass_ctx_obj
@click.argument("name", type=str)
@click.option(
    "--half-life-days",
    type=int,
    default=None,
    help="Half-life for exponential decay in days",
)
@click.option(
    "--lookback-days",
    type=int,
    default=None,
    help="Total lookback period in days",
)
@click.option(
    "--decay-unit-days",
    type=int,
    default=None,
    help="Granularity of decay buckets in days",
)
@click.option(
    "--default-weight",
    type=str,
    default=None,
    help="Default weight for entities",
)
@click.option(
    "--resource-weight",
    "resource_weights",
    type=str,
    multiple=True,
    help="Resource weight in KEY=VALUE format (e.g., cpu=1.0, mem=0.5). Use VALUE=null to remove.",
)
@click.option("--json", "as_json", is_flag=True, help="Output as JSON")
def resource_group_update_spec_cmd(
    ctx: CLIContext,
    name: str,
    half_life_days: int | None,
    lookback_days: int | None,
    decay_unit_days: int | None,
    default_weight: str | None,
    resource_weights: tuple[str, ...],
    as_json: bool,
) -> None:
    """Update resource group fair share specification."""
    from ai.backend.cli.types import ExitCode
    from ai.backend.client.session import Session
    from ai.backend.common.dto.manager.fair_share import ResourceWeightEntryInput

    # Parse resource_weights from "cpu=1.0,mem=null" format
    parsed_weights: list[ResourceWeightEntryInput] | None = None
    if resource_weights:
        parsed_weights = []
        for entry in resource_weights:
            if "=" not in entry:
                ctx.output.print_error(
                    ValueError(
                        f"Invalid resource weight format: {entry}. Expected KEY=VALUE format."
                    )
                )
                sys.exit(ExitCode.FAILURE)
            key, _, value = entry.partition("=")
            weight: Decimal | None = None if value.lower() == "null" else Decimal(value)
            parsed_weights.append(ResourceWeightEntryInput(resource_type=key, weight=weight))

    default_weight_decimal = Decimal(default_weight) if default_weight is not None else None

    with Session() as api_session:
        try:
            response = api_session.FairShare.update_resource_group_fair_share_spec(
                name,
                half_life_days,
                lookback_days,
                decay_unit_days,
                default_weight_decimal,
                parsed_weights,
            )

            if as_json:
                print(json.dumps(response.item.model_dump(mode="json"), indent=2, default=str))
            else:
                spec = response.item
                print("Updated resource group fair share spec successfully.")
                print(f"Resource Group: {name}")
                print(f"Half Life Days: {spec.half_life_days}")
                print(f"Lookback Days: {spec.lookback_days}")
                print(f"Decay Unit Days: {spec.decay_unit_days}")
                print(f"Default Weight: {spec.default_weight}")
                print("Resource Weights:")
                for entry in spec.resource_weights.entries:
                    print(f"  {entry.resource_type}: {entry.quantity}")
        except Exception as e:
            ctx.output.print_error(e)
            sys.exit(ExitCode.FAILURE)


@resource_group.command("spec")
@pass_ctx_obj
@click.argument("name", type=str)
@click.option("--json", "as_json", is_flag=True, help="Output as JSON")
def resource_group_spec_cmd(
    ctx: CLIContext,
    name: str,
    as_json: bool,
) -> None:
    """Get resource group fair share specification."""
    from ai.backend.cli.types import ExitCode
    from ai.backend.client.session import Session

    with Session() as api_session:
        try:
            response = api_session.FairShare.get_resource_group_fair_share_spec(name)

            if as_json:
                print(json.dumps(response.model_dump(mode="json"), indent=2, default=str))
            else:
                print(f"Resource Group: {response.resource_group}")
                spec = response.fair_share_spec
                print(f"Half Life Days: {spec.half_life_days}")
                print(f"Lookback Days: {spec.lookback_days}")
                print(f"Decay Unit Days: {spec.decay_unit_days}")
                print(f"Default Weight: {spec.default_weight}")
                print("Resource Weights:")
                for entry in spec.resource_weights.entries:
                    print(f"  {entry.resource_type}: {entry.quantity}")
        except Exception as e:
            ctx.output.print_error(e)
            sys.exit(ExitCode.FAILURE)


@resource_group.command("specs")
@pass_ctx_obj
@click.option("--json", "as_json", is_flag=True, help="Output as JSON")
def resource_group_specs_cmd(
    ctx: CLIContext,
    as_json: bool,
) -> None:
    """List all resource groups with their fair share specifications."""
    from ai.backend.cli.types import ExitCode
    from ai.backend.client.session import Session

    with Session() as api_session:
        try:
            response = api_session.FairShare.list_resource_group_fair_share_specs()

            if as_json:
                print(
                    json.dumps(
                        [item.model_dump(mode="json") for item in response.items],
                        indent=2,
                        default=str,
                    )
                )
            else:
                items = response.items
                if not items:
                    print("No resource groups found")
                    return
                print(f"Total: {response.total_count}")
                print()
                for item in items:
                    print(f"Resource Group: {item.resource_group}")
                    spec = item.fair_share_spec
                    print(f"  Half Life Days: {spec.half_life_days}")
                    print(f"  Lookback Days: {spec.lookback_days}")
                    print(f"  Decay Unit Days: {spec.decay_unit_days}")
                    print(f"  Default Weight: {spec.default_weight}")
                    print("  Resource Weights:")
                    for entry in spec.resource_weights.entries:
                        print(f"    {entry.resource_type}: {entry.quantity}")
                    print("---")
        except Exception as e:
            ctx.output.print_error(e)
            sys.exit(ExitCode.FAILURE)


# =============================================================================
# RG-Scoped Domain Fair Share Commands (Resource Group Scope)
# =============================================================================


@fair_share.group("rg-domain")
def rg_domain() -> None:
    """Domain fair share operations within resource group scope."""


@rg_domain.command("get")
@pass_ctx_obj
@click.argument("resource_group", type=str)
@click.argument("domain_name", type=str)
@click.option("--json", "as_json", is_flag=True, help="Output as JSON")
def rg_domain_get_cmd(
    ctx: CLIContext,
    resource_group: str,
    domain_name: str,
    as_json: bool,
) -> None:
    """Get domain fair share data within resource group scope."""
    from ai.backend.cli.types import ExitCode
    from ai.backend.client.session import Session

    with Session() as api_session:
        try:
            response = api_session.FairShare.rg_get_domain_fair_share(resource_group, domain_name)
            if response.item is None:
                print("No domain fair share found")
                return

            if as_json:
                print(json.dumps(response.item.model_dump(mode="json"), indent=2, default=str))
            else:
                fs = response.item
                print(f"Resource Group: {fs.resource_group}")
                print(f"Domain: {fs.domain_name}")
                print(f"Weight: {fs.spec.weight}")
                print(f"Fair Share Factor: {fs.calculation_snapshot.fair_share_factor}")
                print(f"Normalized Usage: {fs.calculation_snapshot.normalized_usage}")
                print(
                    f"Lookback: {fs.calculation_snapshot.lookback_start} ~ "
                    f"{fs.calculation_snapshot.lookback_end}"
                )
                print(f"Created: {fs.created_at}")
                print(f"Updated: {fs.updated_at}")
        except Exception as e:
            ctx.output.print_error(e)
            sys.exit(ExitCode.FAILURE)


@rg_domain.command("list")
@pass_ctx_obj
@click.argument("resource_group", type=str)
@click.option("--domain-name", type=str, default=None, help="Filter by domain name")
@click.option("--limit", type=int, default=20, help="Maximum number of records to return")
@click.option("--offset", type=int, default=0, help="Offset for pagination")
@click.option(
    "--order-by",
    type=click.Choice(["fair_share_factor", "domain_name", "created_at"]),
    default="fair_share_factor",
    help="Order by field",
)
@click.option(
    "--order",
    type=click.Choice(["ASC", "DESC"]),
    default="DESC",
    help="Order direction",
)
@click.option("--json", "as_json", is_flag=True, help="Output as JSON")
def rg_domain_list_cmd(
    ctx: CLIContext,
    resource_group: str,
    domain_name: str | None,
    limit: int,
    offset: int,
    order_by: str,
    order: str,
    as_json: bool,
) -> None:
    """List domain fair shares within resource group scope."""
    from ai.backend.cli.types import ExitCode
    from ai.backend.client.session import Session
    from ai.backend.common.dto.manager.fair_share import (
        DomainFairShareFilter,
        DomainFairShareOrder,
        DomainFairShareOrderField,
        OrderDirection,
        SearchDomainFairSharesRequest,
    )
    from ai.backend.common.dto.manager.query import StringFilter

    with Session() as api_session:
        try:
            filter_cond = None
            if domain_name:
                filter_cond = DomainFairShareFilter(
                    domain_name=StringFilter(equals=domain_name),
                )

            order_spec = DomainFairShareOrder(
                field=DomainFairShareOrderField(order_by),
                direction=OrderDirection(order.lower()),
            )

            request = SearchDomainFairSharesRequest(
                filter=filter_cond,
                order=[order_spec],
                limit=limit,
                offset=offset,
            )
            response = api_session.FairShare.rg_search_domain_fair_shares(resource_group, request)

            if as_json:
                print(
                    json.dumps(
                        [fs.model_dump(mode="json") for fs in response.items],
                        indent=2,
                        default=str,
                    )
                )
            else:
                items = response.items
                if not items:
                    print("No domain fair shares found")
                    return
                print(f"Total: {response.pagination.total}")
                print()
                for fs in items:
                    print(f"ID: {fs.id}")
                    print(f"Resource Group: {fs.resource_group}")
                    print(f"Domain: {fs.domain_name}")
                    print(f"Weight: {fs.spec.weight}")
                    print(f"Fair Share Factor: {fs.calculation_snapshot.fair_share_factor}")
                    print(f"Normalized Usage: {fs.calculation_snapshot.normalized_usage}")
                    print(f"Created: {fs.created_at}")
                    print("---")
        except Exception as e:
            ctx.output.print_error(e)
            sys.exit(ExitCode.FAILURE)


# =============================================================================
# RG-Scoped Project Fair Share Commands (Resource Group Scope)
# =============================================================================


@fair_share.group("rg-project")
def rg_project() -> None:
    """Project fair share operations within resource group scope."""


@rg_project.command("get")
@pass_ctx_obj
@click.argument("resource_group", type=str)
@click.argument("domain_name", type=str)
@click.argument("project_id", type=str)
@click.option("--json", "as_json", is_flag=True, help="Output as JSON")
def rg_project_get_cmd(
    ctx: CLIContext,
    resource_group: str,
    domain_name: str,
    project_id: str,
    as_json: bool,
) -> None:
    """Get project fair share data within resource group scope."""
    from ai.backend.cli.types import ExitCode
    from ai.backend.client.session import Session

    with Session() as api_session:
        try:
            response = api_session.FairShare.rg_get_project_fair_share(
                resource_group, domain_name, UUID(project_id)
            )
            if response.item is None:
                print("No project fair share found")
                return

            if as_json:
                print(json.dumps(response.item.model_dump(mode="json"), indent=2, default=str))
            else:
                fs = response.item
                print(f"Resource Group: {fs.resource_group}")
                print(f"Project ID: {fs.project_id}")
                print(f"Domain: {fs.domain_name}")
                print(f"Weight: {fs.spec.weight}")
                print(f"Fair Share Factor: {fs.calculation_snapshot.fair_share_factor}")
                print(f"Normalized Usage: {fs.calculation_snapshot.normalized_usage}")
                print(
                    f"Lookback: {fs.calculation_snapshot.lookback_start} ~ "
                    f"{fs.calculation_snapshot.lookback_end}"
                )
                print(f"Created: {fs.created_at}")
                print(f"Updated: {fs.updated_at}")
        except Exception as e:
            ctx.output.print_error(e)
            sys.exit(ExitCode.FAILURE)


@rg_project.command("list")
@pass_ctx_obj
@click.argument("resource_group", type=str)
@click.argument("domain_name", type=str)
@click.option("--project-id", type=str, default=None, help="Filter by project ID")
@click.option("--limit", type=int, default=20, help="Maximum number of records to return")
@click.option("--offset", type=int, default=0, help="Offset for pagination")
@click.option(
    "--order-by",
    type=click.Choice(["fair_share_factor", "created_at"]),
    default="fair_share_factor",
    help="Order by field",
)
@click.option(
    "--order",
    type=click.Choice(["ASC", "DESC"]),
    default="DESC",
    help="Order direction",
)
@click.option("--json", "as_json", is_flag=True, help="Output as JSON")
def rg_project_list_cmd(
    ctx: CLIContext,
    resource_group: str,
    domain_name: str,
    project_id: str | None,
    limit: int,
    offset: int,
    order_by: str,
    order: str,
    as_json: bool,
) -> None:
    """List project fair shares within resource group scope."""
    from ai.backend.cli.types import ExitCode
    from ai.backend.client.session import Session
    from ai.backend.common.dto.manager.fair_share import (
        OrderDirection,
        ProjectFairShareFilter,
        ProjectFairShareOrder,
        ProjectFairShareOrderField,
        SearchProjectFairSharesRequest,
    )
    from ai.backend.common.dto.manager.query import UUIDFilter

    with Session() as api_session:
        try:
            filter_cond = None
            if project_id:
                filter_cond = ProjectFairShareFilter(
                    project_id=UUIDFilter(equals=UUID(project_id)),
                )

            order_spec = ProjectFairShareOrder(
                field=ProjectFairShareOrderField(order_by),
                direction=OrderDirection(order.lower()),
            )

            request = SearchProjectFairSharesRequest(
                filter=filter_cond,
                order=[order_spec],
                limit=limit,
                offset=offset,
            )
            response = api_session.FairShare.rg_search_project_fair_shares(
                resource_group, domain_name, request
            )

            if as_json:
                print(
                    json.dumps(
                        [fs.model_dump(mode="json") for fs in response.items],
                        indent=2,
                        default=str,
                    )
                )
            else:
                items = response.items
                if not items:
                    print("No project fair shares found")
                    return
                print(f"Total: {response.pagination.total}")
                print()
                for fs in items:
                    print(f"ID: {fs.id}")
                    print(f"Resource Group: {fs.resource_group}")
                    print(f"Project ID: {fs.project_id}")
                    print(f"Domain: {fs.domain_name}")
                    print(f"Weight: {fs.spec.weight}")
                    print(f"Fair Share Factor: {fs.calculation_snapshot.fair_share_factor}")
                    print(f"Normalized Usage: {fs.calculation_snapshot.normalized_usage}")
                    print(f"Created: {fs.created_at}")
                    print("---")
        except Exception as e:
            ctx.output.print_error(e)
            sys.exit(ExitCode.FAILURE)


# =============================================================================
# RG-Scoped User Fair Share Commands (Resource Group Scope)
# =============================================================================


@fair_share.group("rg-user")
def rg_user() -> None:
    """User fair share operations within resource group scope."""


@rg_user.command("get")
@pass_ctx_obj
@click.argument("resource_group", type=str)
@click.argument("domain_name", type=str)
@click.argument("project_id", type=str)
@click.argument("user_uuid", type=str)
@click.option("--json", "as_json", is_flag=True, help="Output as JSON")
def rg_user_get_cmd(
    ctx: CLIContext,
    resource_group: str,
    domain_name: str,
    project_id: str,
    user_uuid: str,
    as_json: bool,
) -> None:
    """Get user fair share data within resource group scope."""
    from ai.backend.cli.types import ExitCode
    from ai.backend.client.session import Session

    with Session() as api_session:
        try:
            response = api_session.FairShare.rg_get_user_fair_share(
                resource_group, domain_name, UUID(project_id), UUID(user_uuid)
            )
            if response.item is None:
                print("No user fair share found")
                return

            if as_json:
                print(json.dumps(response.item.model_dump(mode="json"), indent=2, default=str))
            else:
                fs = response.item
                print(f"Resource Group: {fs.resource_group}")
                print(f"User UUID: {fs.user_uuid}")
                print(f"Project ID: {fs.project_id}")
                print(f"Domain: {fs.domain_name}")
                print(f"Weight: {fs.spec.weight}")
                print(f"Fair Share Factor: {fs.calculation_snapshot.fair_share_factor}")
                print(f"Normalized Usage: {fs.calculation_snapshot.normalized_usage}")
                print(
                    f"Lookback: {fs.calculation_snapshot.lookback_start} ~ "
                    f"{fs.calculation_snapshot.lookback_end}"
                )
                print(f"Created: {fs.created_at}")
                print(f"Updated: {fs.updated_at}")
        except Exception as e:
            ctx.output.print_error(e)
            sys.exit(ExitCode.FAILURE)


@rg_user.command("list")
@pass_ctx_obj
@click.argument("resource_group", type=str)
@click.argument("domain_name", type=str)
@click.argument("project_id", type=str)
@click.option("--user-uuid", type=str, default=None, help="Filter by user UUID")
@click.option("--limit", type=int, default=20, help="Maximum number of records to return")
@click.option("--offset", type=int, default=0, help="Offset for pagination")
@click.option(
    "--order-by",
    type=click.Choice(["fair_share_factor", "created_at"]),
    default="fair_share_factor",
    help="Order by field",
)
@click.option(
    "--order",
    type=click.Choice(["ASC", "DESC"]),
    default="DESC",
    help="Order direction",
)
@click.option("--json", "as_json", is_flag=True, help="Output as JSON")
def rg_user_list_cmd(
    ctx: CLIContext,
    resource_group: str,
    domain_name: str,
    project_id: str,
    user_uuid: str | None,
    limit: int,
    offset: int,
    order_by: str,
    order: str,
    as_json: bool,
) -> None:
    """List user fair shares within resource group scope."""
    from ai.backend.cli.types import ExitCode
    from ai.backend.client.session import Session
    from ai.backend.common.dto.manager.fair_share import (
        OrderDirection,
        SearchUserFairSharesRequest,
        UserFairShareFilter,
        UserFairShareOrder,
        UserFairShareOrderField,
    )
    from ai.backend.common.dto.manager.query import UUIDFilter

    with Session() as api_session:
        try:
            filter_cond = None
            if user_uuid:
                filter_cond = UserFairShareFilter(
                    user_uuid=UUIDFilter(equals=UUID(user_uuid)),
                )

            order_spec = UserFairShareOrder(
                field=UserFairShareOrderField(order_by),
                direction=OrderDirection(order.lower()),
            )

            request = SearchUserFairSharesRequest(
                filter=filter_cond,
                order=[order_spec],
                limit=limit,
                offset=offset,
            )
            response = api_session.FairShare.rg_search_user_fair_shares(
                resource_group, domain_name, UUID(project_id), request
            )

            if as_json:
                print(
                    json.dumps(
                        [fs.model_dump(mode="json") for fs in response.items],
                        indent=2,
                        default=str,
                    )
                )
            else:
                items = response.items
                if not items:
                    print("No user fair shares found")
                    return
                print(f"Total: {response.pagination.total}")
                print()
                for fs in items:
                    print(f"ID: {fs.id}")
                    print(f"Resource Group: {fs.resource_group}")
                    print(f"User UUID: {fs.user_uuid}")
                    print(f"Project ID: {fs.project_id}")
                    print(f"Domain: {fs.domain_name}")
                    print(f"Weight: {fs.spec.weight}")
                    print(f"Fair Share Factor: {fs.calculation_snapshot.fair_share_factor}")
                    print(f"Normalized Usage: {fs.calculation_snapshot.normalized_usage}")
                    print(f"Created: {fs.created_at}")
                    print("---")
        except Exception as e:
            ctx.output.print_error(e)
            sys.exit(ExitCode.FAILURE)


# =============================================================================
# RG-Scoped Domain Usage Bucket Commands
# =============================================================================


@fair_share.group("rg-domain-usage")
def rg_domain_usage() -> None:
    """Domain usage buckets within resource group scope."""


@rg_domain_usage.command("list")
@pass_ctx_obj
@click.argument("resource_group", type=str)
@click.option("--domain-name", type=str, default=None, help="Filter by domain name")
@click.option(
    "--period-start-after",
    type=click.DateTime(formats=["%Y-%m-%d"]),
    default=None,
    help="Filter buckets with period_start >= this date (YYYY-MM-DD)",
)
@click.option(
    "--period-start-before",
    type=click.DateTime(formats=["%Y-%m-%d"]),
    default=None,
    help="Filter buckets with period_start <= this date (YYYY-MM-DD)",
)
@click.option("--limit", type=int, default=20, help="Maximum number of records to return")
@click.option("--offset", type=int, default=0, help="Offset for pagination")
@click.option("--json", "as_json", is_flag=True, help="Output as JSON")
def rg_domain_usage_list_cmd(
    ctx: CLIContext,
    resource_group: str,
    domain_name: str | None,
    period_start_after: click.DateTime | None,
    period_start_before: click.DateTime | None,
    limit: int,
    offset: int,
    as_json: bool,
) -> None:
    """List domain usage buckets within resource group scope."""
    from ai.backend.cli.types import ExitCode
    from ai.backend.client.session import Session
    from ai.backend.common.dto.manager.fair_share import (
        DomainUsageBucketFilter,
        DomainUsageBucketOrder,
        DomainUsageBucketOrderField,
        OrderDirection,
        SearchDomainUsageBucketsRequest,
    )
    from ai.backend.common.dto.manager.query import DateRangeFilter, StringFilter

    with Session() as api_session:
        try:
            filter_cond = None
            if domain_name or period_start_after or period_start_before:
                period_start_filter = None
                if period_start_after or period_start_before:
                    period_start_filter = DateRangeFilter(
                        after=period_start_after.date() if period_start_after else None,
                        before=period_start_before.date() if period_start_before else None,
                    )

                filter_cond = DomainUsageBucketFilter(
                    domain_name=StringFilter(equals=domain_name) if domain_name else None,
                    period_start=period_start_filter,
                )

            order_spec = DomainUsageBucketOrder(
                field=DomainUsageBucketOrderField.PERIOD_START,
                direction=OrderDirection.DESC,
            )

            request = SearchDomainUsageBucketsRequest(
                filter=filter_cond,
                order=[order_spec],
                limit=limit,
                offset=offset,
            )
            response = api_session.FairShare.rg_search_domain_usage_buckets(resource_group, request)

            if as_json:
                print(
                    json.dumps(
                        [item.model_dump(mode="json") for item in response.items],
                        indent=2,
                        default=str,
                    )
                )
            else:
                items = response.items
                if not items:
                    print("No domain usage buckets found")
                    return
                print(f"Total: {response.pagination.total}")
                print()
                for bucket in items:
                    print(f"Period: {bucket.metadata.period_start} ~ {bucket.metadata.period_end}")
                    print(f"Domain: {bucket.domain_name}")
                    print(f"Resource Group: {bucket.resource_group}")
                    print("Resource Usage:")
                    for entry in bucket.resource_usage.entries:
                        print(f"  {entry.resource_type}: {entry.quantity}")
                    print(f"Created: {bucket.metadata.created_at}")
                    print("---")
        except Exception as e:
            ctx.output.print_error(e)
            sys.exit(ExitCode.FAILURE)


# =============================================================================
# RG-Scoped Project Usage Bucket Commands
# =============================================================================


@fair_share.group("rg-project-usage")
def rg_project_usage() -> None:
    """Project usage buckets within resource group scope."""


@rg_project_usage.command("list")
@pass_ctx_obj
@click.argument("resource_group", type=str)
@click.argument("domain_name", type=str)
@click.option("--project-id", type=str, default=None, help="Filter by project ID (UUID)")
@click.option(
    "--period-start-after",
    type=click.DateTime(formats=["%Y-%m-%d"]),
    default=None,
    help="Filter buckets with period_start >= this date (YYYY-MM-DD)",
)
@click.option(
    "--period-start-before",
    type=click.DateTime(formats=["%Y-%m-%d"]),
    default=None,
    help="Filter buckets with period_start <= this date (YYYY-MM-DD)",
)
@click.option("--limit", type=int, default=20, help="Maximum number of records to return")
@click.option("--offset", type=int, default=0, help="Offset for pagination")
@click.option("--json", "as_json", is_flag=True, help="Output as JSON")
def rg_project_usage_list_cmd(
    ctx: CLIContext,
    resource_group: str,
    domain_name: str,
    project_id: str | None,
    period_start_after: click.DateTime | None,
    period_start_before: click.DateTime | None,
    limit: int,
    offset: int,
    as_json: bool,
) -> None:
    """List project usage buckets within resource group scope."""
    from ai.backend.cli.types import ExitCode
    from ai.backend.client.session import Session
    from ai.backend.common.dto.manager.fair_share import (
        OrderDirection,
        ProjectUsageBucketFilter,
        ProjectUsageBucketOrder,
        ProjectUsageBucketOrderField,
        SearchProjectUsageBucketsRequest,
    )
    from ai.backend.common.dto.manager.query import DateRangeFilter, UUIDFilter

    with Session() as api_session:
        try:
            filter_cond = None
            if project_id or period_start_after or period_start_before:
                period_start_filter = None
                if period_start_after or period_start_before:
                    period_start_filter = DateRangeFilter(
                        after=period_start_after.date() if period_start_after else None,
                        before=period_start_before.date() if period_start_before else None,
                    )

                filter_cond = ProjectUsageBucketFilter(
                    project_id=UUIDFilter(equals=UUID(project_id)) if project_id else None,
                    period_start=period_start_filter,
                )

            order_spec = ProjectUsageBucketOrder(
                field=ProjectUsageBucketOrderField.PERIOD_START,
                direction=OrderDirection.DESC,
            )

            request = SearchProjectUsageBucketsRequest(
                filter=filter_cond,
                order=[order_spec],
                limit=limit,
                offset=offset,
            )
            response = api_session.FairShare.rg_search_project_usage_buckets(
                resource_group, domain_name, request
            )

            if as_json:
                print(
                    json.dumps(
                        [item.model_dump(mode="json") for item in response.items],
                        indent=2,
                        default=str,
                    )
                )
            else:
                items = response.items
                if not items:
                    print("No project usage buckets found")
                    return
                print(f"Total: {response.pagination.total}")
                print()
                for bucket in items:
                    print(f"Period: {bucket.metadata.period_start} ~ {bucket.metadata.period_end}")
                    print(f"Project ID: {bucket.project_id}")
                    print(f"Domain: {bucket.domain_name}")
                    print(f"Resource Group: {bucket.resource_group}")
                    print("Resource Usage:")
                    for entry in bucket.resource_usage.entries:
                        print(f"  {entry.resource_type}: {entry.quantity}")
                    print(f"Created: {bucket.metadata.created_at}")
                    print("---")
        except Exception as e:
            ctx.output.print_error(e)
            sys.exit(ExitCode.FAILURE)


# =============================================================================
# RG-Scoped User Usage Bucket Commands
# =============================================================================


@fair_share.group("rg-user-usage")
def rg_user_usage() -> None:
    """User usage buckets within resource group scope."""


@rg_user_usage.command("list")
@pass_ctx_obj
@click.argument("resource_group", type=str)
@click.argument("domain_name", type=str)
@click.argument("project_id", type=str)
@click.option("--user-uuid", type=str, default=None, help="Filter by user UUID")
@click.option(
    "--period-start-after",
    type=click.DateTime(formats=["%Y-%m-%d"]),
    default=None,
    help="Filter buckets with period_start >= this date (YYYY-MM-DD)",
)
@click.option(
    "--period-start-before",
    type=click.DateTime(formats=["%Y-%m-%d"]),
    default=None,
    help="Filter buckets with period_start <= this date (YYYY-MM-DD)",
)
@click.option("--limit", type=int, default=20, help="Maximum number of records to return")
@click.option("--offset", type=int, default=0, help="Offset for pagination")
@click.option("--json", "as_json", is_flag=True, help="Output as JSON")
def rg_user_usage_list_cmd(
    ctx: CLIContext,
    resource_group: str,
    domain_name: str,
    project_id: str,
    user_uuid: str | None,
    period_start_after: click.DateTime | None,
    period_start_before: click.DateTime | None,
    limit: int,
    offset: int,
    as_json: bool,
) -> None:
    """List user usage buckets within resource group scope."""
    from ai.backend.cli.types import ExitCode
    from ai.backend.client.session import Session
    from ai.backend.common.dto.manager.fair_share import (
        OrderDirection,
        SearchUserUsageBucketsRequest,
        UserUsageBucketFilter,
        UserUsageBucketOrder,
        UserUsageBucketOrderField,
    )
    from ai.backend.common.dto.manager.query import DateRangeFilter, UUIDFilter

    with Session() as api_session:
        try:
            filter_cond = None
            if user_uuid or period_start_after or period_start_before:
                period_start_filter = None
                if period_start_after or period_start_before:
                    period_start_filter = DateRangeFilter(
                        after=period_start_after.date() if period_start_after else None,
                        before=period_start_before.date() if period_start_before else None,
                    )

                filter_cond = UserUsageBucketFilter(
                    user_uuid=UUIDFilter(equals=UUID(user_uuid)) if user_uuid else None,
                    period_start=period_start_filter,
                )

            order_spec = UserUsageBucketOrder(
                field=UserUsageBucketOrderField.PERIOD_START,
                direction=OrderDirection.DESC,
            )

            request = SearchUserUsageBucketsRequest(
                filter=filter_cond,
                order=[order_spec],
                limit=limit,
                offset=offset,
            )
            response = api_session.FairShare.rg_search_user_usage_buckets(
                resource_group, domain_name, UUID(project_id), request
            )

            if as_json:
                print(
                    json.dumps(
                        [item.model_dump(mode="json") for item in response.items],
                        indent=2,
                        default=str,
                    )
                )
            else:
                items = response.items
                if not items:
                    print("No user usage buckets found")
                    return
                print(f"Total: {response.pagination.total}")
                print()
                for bucket in items:
                    print(f"Period: {bucket.metadata.period_start} ~ {bucket.metadata.period_end}")
                    print(f"User UUID: {bucket.user_uuid}")
                    print(f"Project ID: {bucket.project_id}")
                    print(f"Domain: {bucket.domain_name}")
                    print(f"Resource Group: {bucket.resource_group}")
                    print("Resource Usage:")
                    for entry in bucket.resource_usage.entries:
                        print(f"  {entry.resource_type}: {entry.quantity}")
                    print(f"Created: {bucket.metadata.created_at}")
                    print("---")
        except Exception as e:
            ctx.output.print_error(e)
            sys.exit(ExitCode.FAILURE)



"""Client SDK functions for scheduling history."""

from __future__ import annotations

from ai.backend.client.request import Request
from ai.backend.common.dto.manager.scheduling_history import (
    ListDeploymentHistoryResponse,
    ListRouteHistoryResponse,
    ListSessionHistoryResponse,
    SearchDeploymentHistoryRequest,
    SearchRouteHistoryRequest,
    SearchSessionHistoryRequest,
)

from .base import BaseFunction, api_function

__all__ = ("SchedulingHistory",)


class SchedulingHistory(BaseFunction):
    """
    Provides functions to interact with scheduling history.
    Supports session, deployment, and route scheduling history.
    Requires superadmin privileges.
    """

    @api_function
    @classmethod
    async def list_session_history(
        cls,
        request: SearchSessionHistoryRequest,
    ) -> ListSessionHistoryResponse:
        """
        Search session scheduling history.

        :param request: Session history search request
        :returns: List of session scheduling history records
        """
        rqst = Request("POST", "/scheduling-history/sessions/search")
        rqst.set_json(request.model_dump(mode="json", exclude_none=True))
        async with rqst.fetch() as resp:
            data = await resp.json()
            return ListSessionHistoryResponse.model_validate(data)

    @api_function
    @classmethod
    async def list_deployment_history(
        cls,
        request: SearchDeploymentHistoryRequest,
    ) -> ListDeploymentHistoryResponse:
        """
        Search deployment scheduling history.

        :param request: Deployment history search request
        :returns: List of deployment history records
        """
        rqst = Request("POST", "/scheduling-history/deployments/search")
        rqst.set_json(request.model_dump(mode="json", exclude_none=True))
        async with rqst.fetch() as resp:
            data = await resp.json()
            return ListDeploymentHistoryResponse.model_validate(data)

    @api_function
    @classmethod
    async def list_route_history(
        cls,
        request: SearchRouteHistoryRequest,
    ) -> ListRouteHistoryResponse:
        """
        Search route scheduling history.

        :param request: Route history search request
        :returns: List of route history records
        """
        rqst = Request("POST", "/scheduling-history/routes/search")
        rqst.set_json(request.model_dump(mode="json", exclude_none=True))
        async with rqst.fetch() as resp:
            data = await resp.json()
            return ListRouteHistoryResponse.model_validate(data)



import decimal
import sys
from collections.abc import Iterable
from typing import Any
from uuid import UUID

import click

from ai.backend.cli.params import OptionalType
from ai.backend.cli.types import ExitCode, Undefined, undefined
from ai.backend.client.cli.extensions import pass_ctx_obj
from ai.backend.client.cli.service import get_service_id
from ai.backend.client.cli.types import CLIContext
from ai.backend.client.exceptions import BackendAPIError
from ai.backend.client.func.service_auto_scaling_rule import _default_fields as _default_get_fields
from ai.backend.client.output.fields import service_auto_scaling_rule_fields
from ai.backend.client.session import Session
from ai.backend.common.types import AutoScalingMetricComparator, AutoScalingMetricSource

from .pretty import print_done
from .service import service

_default_list_fields = (
    service_auto_scaling_rule_fields["id"],
    service_auto_scaling_rule_fields["metric_source"],
    service_auto_scaling_rule_fields["metric_name"],
    service_auto_scaling_rule_fields["comparator"],
    service_auto_scaling_rule_fields["threshold"],
)


@service.group()
def auto_scaling_rule() -> None:
    """Set of model service auto-scaling rule operations"""


@auto_scaling_rule.command()
@pass_ctx_obj
@click.argument("service", type=str, metavar="SERVICE_NAME_OR_ID")
@click.option(
    "--metric-source",
    type=click.Choice([*AutoScalingMetricSource], case_sensitive=False),
    required=True,
)
@click.option("--metric-name", type=str, required=True)
@click.option("--threshold", type=str, required=True)
@click.option(
    "--comparator",
    type=click.Choice([*AutoScalingMetricComparator], case_sensitive=False),
    required=True,
)
@click.option("--step-size", type=int, required=True)
@click.option("--cooldown-seconds", type=int, required=True)
@click.option("--min-replicas", type=int)
@click.option("--max-replicas", type=int)
def create(
    ctx: CLIContext,
    service: str,
    *,
    metric_source: AutoScalingMetricSource,
    metric_name: str,
    threshold: str,
    comparator: AutoScalingMetricComparator,
    step_size: int,
    cooldown_seconds: int,
    min_replicas: int | None = None,
    max_replicas: int | None = None,
) -> None:
    """Create a new auto-scaling rule."""

    with Session() as session:
        try:
            _threshold = decimal.Decimal(threshold)
        except decimal.InvalidOperation:
            ctx.output.print_fail(f"{threshold} is not a valid Decimal")
            sys.exit(ExitCode.FAILURE)

        try:
            service_id = get_service_id(session, service)
            rule = session.ServiceAutoScalingRule.create(
                service_id,
                metric_source,
                metric_name,
                _threshold,
                comparator,
                step_size,
                cooldown_seconds,
                min_replicas=min_replicas,
                max_replicas=max_replicas,
            )
            print_done(f"Auto-scaling Rule (ID {rule.rule_id}) created.")
        except Exception as e:
            ctx.output.print_error(e)
            sys.exit(ExitCode.FAILURE)


@auto_scaling_rule.command()
@pass_ctx_obj
@click.argument("service", type=str, metavar="SERVICE_NAME_OR_ID")
@click.option(
    "-f",
    "--format",
    default=None,
    help="Display only specified fields.  When specifying multiple fields separate them with comma (,).",
)
@click.option("--filter", "filter_", default=None, help="Set the query filter expression.")
@click.option("--order", default=None, help="Set the query ordering expression.")
@click.option("--offset", default=0, help="The index of the current page start for pagination.")
@click.option("--limit", type=int, default=None, help="The page size for pagination.")
def list(
    ctx: CLIContext,
    service: str,
    format: str | None,
    filter_: str | None,
    order: str | None,
    offset: int,
    limit: int | None,
) -> None:
    """List all set auto-scaling rules for given model service."""

    if format:
        try:
            fields = [service_auto_scaling_rule_fields[f.strip()] for f in format.split(",")]
        except KeyError as e:
            ctx.output.print_fail(f"Field {e!s} not found")
            sys.exit(ExitCode.FAILURE)
    else:
        fields = None
    with Session() as session:
        service_id = get_service_id(session, service)

        try:
            fetch_func = lambda pg_offset, pg_size: session.ServiceAutoScalingRule.paginated_list(
                service_id,
                page_offset=pg_offset,
                page_size=pg_size,
                filter=filter_,
                order=order,
                fields=fields,
            )
            ctx.output.print_paginated_list(
                fetch_func,
                initial_page_offset=offset,
                page_size=limit,
            )
        except Exception as e:
            ctx.output.print_error(e)
            sys.exit(ExitCode.FAILURE)


@auto_scaling_rule.command()
@pass_ctx_obj
@click.argument("rule", type=click.UUID, metavar="RULE_ID")
@click.option(
    "-f",
    "--format",
    default=None,
    help="Display only specified fields.  When specifying multiple fields separate them with comma (,).",
)
def get(ctx: CLIContext, rule: UUID, format: str) -> None:
    """Prints attributes of the given auto-scaling rule."""
    fields: Iterable[Any]
    if format:
        try:
            fields = [service_auto_scaling_rule_fields[f.strip()] for f in format.split(",")]
        except KeyError as e:
            ctx.output.print_fail(f"Field {e!s} not found")
            sys.exit(ExitCode.FAILURE)
    else:
        fields = _default_get_fields

    with Session() as session:
        try:
            rule_instance = session.ServiceAutoScalingRule(rule).get(fields=fields)
        except (ValueError, BackendAPIError):
            ctx.output.print_fail(f"Rule {rule!r} not found.")
            sys.exit(ExitCode.FAILURE)

        ctx.output.print_item(rule_instance, fields)


@auto_scaling_rule.command()
@pass_ctx_obj
@click.argument("rule", type=click.UUID, metavar="RULE_ID")
@click.option(
    "--metric-source",
    type=OptionalType(click.Choice([*AutoScalingMetricSource], case_sensitive=False)),
    default=undefined,
)
@click.option("--metric-name", type=OptionalType(str), default=undefined)
@click.option("--threshold", type=OptionalType(str), default=undefined)
@click.option(
    "--comparator",
    type=OptionalType(click.Choice([*AutoScalingMetricComparator], case_sensitive=False)),
    default=undefined,
)
@click.option("--step-size", type=OptionalType(int), default=undefined)
@click.option("--cooldown-seconds", type=OptionalType(int), default=undefined)
@click.option(
    "--min-replicas",
    type=OptionalType(int),
    help="Set as -1 to remove min_replicas restriction.",
    default=undefined,
)
@click.option(
    "--max-replicas",
    type=OptionalType(int),
    help="Set as -1 to remove max_replicas restriction.",
    default=undefined,
)
def update(
    ctx: CLIContext,
    rule: UUID,
    *,
    metric_source: str | Undefined,
    metric_name: str | Undefined,
    threshold: str | Undefined,
    comparator: str | Undefined,
    step_size: int | Undefined,
    cooldown_seconds: int | Undefined,
    min_replicas: int | None | Undefined,
    max_replicas: int | None | Undefined,
) -> None:
    """Update attributes of the given auto-scaling rule."""
    with Session() as session:
        try:
            _threshold = decimal.Decimal(threshold) if threshold != undefined else undefined
        except decimal.InvalidOperation:
            ctx.output.print_fail(f"{threshold} is not a valid Decimal")
            sys.exit(ExitCode.FAILURE)

        if min_replicas == -1:
            min_replicas = None
        if max_replicas == -1:
            max_replicas = None

        try:
            rule_instance = session.ServiceAutoScalingRule(rule)
            _ = rule_instance.get()
            _ = rule_instance.update(
                metric_source=metric_source,
                metric_name=metric_name,
                threshold=_threshold,
                comparator=comparator,
                step_size=step_size,
                cooldown_seconds=cooldown_seconds,
                min_replicas=min_replicas,
                max_replicas=max_replicas,
            )
            print_done(f"Auto-scaling Rule (ID {rule_instance.rule_id}) updated.")
        except BackendAPIError as e:
            ctx.output.print_fail(e.data["title"])
            sys.exit(ExitCode.FAILURE)


@auto_scaling_rule.command()
@pass_ctx_obj
@click.argument("rule", type=click.UUID, metavar="RULE_ID")
def delete(ctx: CLIContext, rule: UUID) -> None:
    """Remove the given auto-scaling rule."""
    with Session() as session:
        rule_instance = session.ServiceAutoScalingRule(rule)
        try:
            _ = rule_instance.get(fields=[service_auto_scaling_rule_fields["id"]])
            _ = rule_instance.delete()
            print_done(f"Autosscaling rule {rule_instance.rule_id} has been deleted.")
        except BackendAPIError as e:
            ctx.output.print_fail(f"Failed to delete rule {rule_instance.rule_id}:")
            ctx.output.print_error(e)
            sys.exit(ExitCode.FAILURE)



"""Client SDK functions for resource usage history."""

from __future__ import annotations

from uuid import UUID

from ai.backend.client.request import Request
from ai.backend.common.dto.manager.fair_share import (
    SearchDomainUsageBucketsRequest,
    SearchDomainUsageBucketsResponse,
    SearchProjectUsageBucketsRequest,
    SearchProjectUsageBucketsResponse,
    SearchUserUsageBucketsRequest,
    SearchUserUsageBucketsResponse,
)

from .base import BaseFunction, api_function

__all__ = ("ResourceUsage",)


class ResourceUsage(BaseFunction):
    """
    Provides functions to interact with resource usage history data.
    Supports domain, project, and user usage buckets.
    Requires superadmin privileges.
    """

    # Domain Usage Buckets

    @api_function
    @classmethod
    async def search_domain_usage_buckets(
        cls,
        request: SearchDomainUsageBucketsRequest,
    ) -> SearchDomainUsageBucketsResponse:
        """
        Search domain usage buckets.

        :param request: Domain usage bucket search request
        :returns: List of domain usage buckets
        """
        rqst = Request("POST", "/fair-share/usage-buckets/domains/search")
        rqst.set_json(request.model_dump(mode="json", exclude_none=True))
        async with rqst.fetch() as resp:
            data = await resp.json()
            return SearchDomainUsageBucketsResponse.model_validate(data)

    # Project Usage Buckets

    @api_function
    @classmethod
    async def search_project_usage_buckets(
        cls,
        request: SearchProjectUsageBucketsRequest,
    ) -> SearchProjectUsageBucketsResponse:
        """
        Search project usage buckets.

        :param request: Project usage bucket search request
        :returns: List of project usage buckets
        """
        rqst = Request("POST", "/fair-share/usage-buckets/projects/search")
        rqst.set_json(request.model_dump(mode="json", exclude_none=True))
        async with rqst.fetch() as resp:
            data = await resp.json()
            return SearchProjectUsageBucketsResponse.model_validate(data)

    # User Usage Buckets

    @api_function
    @classmethod
    async def search_user_usage_buckets(
        cls,
        request: SearchUserUsageBucketsRequest,
    ) -> SearchUserUsageBucketsResponse:
        """
        Search user usage buckets.

        :param request: User usage bucket search request
        :returns: List of user usage buckets
        """
        rqst = Request("POST", "/fair-share/usage-buckets/users/search")
        rqst.set_json(request.model_dump(mode="json", exclude_none=True))
        async with rqst.fetch() as resp:
            data = await resp.json()
            return SearchUserUsageBucketsResponse.model_validate(data)

    # RG-Scoped Usage Bucket Search Methods

    @api_function
    @classmethod
    async def rg_search_domain_usage_buckets(
        cls,
        resource_group: str,
        request: SearchDomainUsageBucketsRequest,
    ) -> SearchDomainUsageBucketsResponse:
        """
        Search domain usage buckets within resource group scope.

        :param resource_group: Resource group name
        :param request: Domain usage bucket search request
        :returns: List of domain usage buckets
        """
        rqst = Request("POST", f"/fair-share/rg/{resource_group}/usage-buckets/domains/search")
        rqst.set_json(request.model_dump(mode="json", exclude_none=True))
        async with rqst.fetch() as resp:
            data = await resp.json()
            return SearchDomainUsageBucketsResponse.model_validate(data)

    @api_function
    @classmethod
    async def rg_search_project_usage_buckets(
        cls,
        resource_group: str,
        domain_name: str,
        request: SearchProjectUsageBucketsRequest,
    ) -> SearchProjectUsageBucketsResponse:
        """
        Search project usage buckets within resource group and domain scope.

        :param resource_group: Resource group name
        :param domain_name: Domain name
        :param request: Project usage bucket search request
        :returns: List of project usage buckets
        """
        rqst = Request(
            "POST",
            f"/fair-share/rg/{resource_group}/domains/{domain_name}/usage-buckets/projects/search",
        )
        rqst.set_json(request.model_dump(mode="json", exclude_none=True))
        async with rqst.fetch() as resp:
            data = await resp.json()
            return SearchProjectUsageBucketsResponse.model_validate(data)

    @api_function
    @classmethod
    async def rg_search_user_usage_buckets(
        cls,
        resource_group: str,
        domain_name: str,
        project_id: UUID,
        request: SearchUserUsageBucketsRequest,
    ) -> SearchUserUsageBucketsResponse:
        """
        Search user usage buckets within resource group, domain, and project scope.

        :param resource_group: Resource group name
        :param domain_name: Domain name
        :param project_id: Project ID
        :param request: User usage bucket search request
        :returns: List of user usage buckets
        """
        rqst = Request(
            "POST",
            f"/fair-share/rg/{resource_group}/domains/{domain_name}/projects/{project_id}/usage-buckets/users/search",
        )
        rqst.set_json(request.model_dump(mode="json", exclude_none=True))
        async with rqst.fetch() as resp:
            data = await resp.json()
            return SearchUserUsageBucketsResponse.model_validate(data)



from collections.abc import Mapping, Sequence
from typing import Any, Literal, cast
from uuid import UUID
from warnings import deprecated

from faker import Faker

from ai.backend.client.exceptions import BackendClientError
from ai.backend.client.output.fields import service_fields
from ai.backend.client.output.types import FieldSpec, PaginatedResult
from ai.backend.client.pagination import fetch_paginated_result
from ai.backend.client.request import Request
from ai.backend.client.session import api_session
from ai.backend.client.utils import dedent as _d
from ai.backend.common.arch import DEFAULT_IMAGE_ARCH
from ai.backend.common.typed_validators import SESSION_NAME_MAX_LENGTH
from ai.backend.common.types import RuntimeVariant

from .base import BaseFunction, api_function

__all__ = ("Service",)

_default_fields: Sequence[FieldSpec] = (
    service_fields["endpoint_id"],
    service_fields["name"],
    service_fields["image_object"],
    service_fields["replicas"],
    service_fields["routings"],
    service_fields["url"],
    service_fields["session_owner"],
    service_fields["open_to_public"],
)


class Service(BaseFunction):
    id: UUID

    @api_function
    @classmethod
    @deprecated("Use paginated_list() instead of this method unless you set the name filter.")
    async def list(cls, name: str | None = None) -> list[dict[str, Any]]:
        params = {}
        if name:
            params["name"] = name
        rqst = Request("GET", "/services", params=params)
        async with rqst.fetch() as resp:
            result: dict[str, Any] = await resp.json()

            return cast(list[dict[str, Any]], result)

    @api_function
    @classmethod
    async def paginated_list(
        cls,
        *,
        fields: Sequence[FieldSpec] = _default_fields,
        page_offset: int = 0,
        page_size: int = 20,
        filter: str | None = None,
        order: str | None = None,
    ) -> PaginatedResult[dict[str, Any]]:
        return await fetch_paginated_result(
            "endpoint_list",
            {
                "filter": (filter, "String"),
                "order": (order, "String"),
            },
            fields,
            page_size=page_size,
            page_offset=page_offset,
        )

    @api_function
    @classmethod
    async def detail(
        cls,
        service_id: str,
        fields: Sequence[FieldSpec] = _default_fields,
    ) -> Sequence[dict[str, Any]]:
        query = _d("""
            query($endpoint_id: UUID!) {
                endpoint(endpoint_id: $endpoint_id) { $fields }
            }
        """)
        query = query.replace("$fields", " ".join(f.field_ref for f in fields))
        variables = {"endpoint_id": service_id}
        data = await api_session.get().Admin._query(query, variables)
        return cast(Sequence[dict[str, Any]], data["endpoint"])

    @api_function
    @classmethod
    async def create(
        cls,
        image: str,
        model_id_or_name: str,
        initial_session_count: int,
        *,
        resources: Mapping[str, str | int],
        resource_opts: Mapping[str, str | int],
        domain_name: str,
        group_name: str,
        scaling_group: str,
        extra_mounts: Sequence[str] | None = None,
        extra_mount_map: Mapping[str, str] | None = None,
        extra_mount_options: Mapping[str, Mapping[str, str]] | None = None,
        service_name: str | None = None,
        model_version: str | None = None,
        _dependencies: Sequence[str] | None = None,
        model_mount_destination: str | None = None,
        envs: Mapping[str, str] | None = None,
        startup_command: str | None = None,
        cluster_size: int = 1,
        cluster_mode: Literal["single-node", "multi-node"] = "single-node",
        bootstrap_script: str | None = None,
        tag: str | None = None,
        architecture: str | None = DEFAULT_IMAGE_ARCH,
        owner_access_key: str | None = None,
        model_definition_path: str | None = None,
        expose_to_public: bool = False,
        runtime_variant: RuntimeVariant | None = None,
    ) -> dict[str, Any]:
        """
        Creates an inference service.

        :param image: The image name and tag for the infernence session.
            Example: ``python:3.6-ubuntu``.
            Check out the full list of available images in your server using (TODO:
            new API).
        :param service_name: A client-side (user-defined) identifier to distinguish the session among currently
            running sessions.
            It may be used to seamlessly reuse the session already created.
        :param initial_session_count: Number of sessions to be started along with
            service initiation.
        :param mounts: The ID of model type vFolder which contains model files required
            to start inference session.
        :param model_mount_destination: Path inside the container to mount model vFolder,
            defaults to /models
        :param envs: The environment variables which always bypasses the jail policy.
        :param resources: The resource specification. (TODO: details)
        :param cluster_size: The number of containers in this compute session.
            Must be at least 1.
        :param cluster_mode: Set the clustering mode whether to use distributed
            nodes or a single node to spawn multiple containers for the new session.
        :param tag: An optional string to annotate extra information.
        :param owner: An optional access key that owns the created session. (Only
            available to administrators)
        :param model_definition_path: Relative path to model definition file. Defaults to `model-definition.yaml`.
        :param expose_to_public: Visibility of API Endpoint which serves inference workload.
            If set to true, no authentication will be required to access the endpoint.
        :param runtime_variant: The runtime variant to use for the service.

        :returns: The :class:`ComputeSession` instance.
        """
        extra_mounts = extra_mounts or []
        extra_mount_map = extra_mount_map or {}
        extra_mount_options = extra_mount_options or {}

        if service_name is None:
            faker = Faker()
            service_name = f"bai-serve-{faker.user_name()}"[:SESSION_NAME_MAX_LENGTH]

        extra_mount_body = {}
        if extra_mounts:
            vfolder_id_to_name: dict[UUID, str] = {}
            vfolder_name_to_id: dict[str, UUID] = {}

            rqst = Request("GET", "/folders")
            async with rqst.fetch() as resp:
                body = await resp.json()
                for folder_info in body:
                    vfolder_id_to_name[UUID(folder_info["id"])] = folder_info["name"]
                    vfolder_name_to_id[folder_info["name"]] = UUID(folder_info["id"])

            for mount in extra_mounts:
                try:
                    vfolder_id = UUID(mount)
                    if vfolder_id not in vfolder_id_to_name:
                        raise BackendClientError(f"VFolder (id: {vfolder_id}) not found")
                except ValueError as e:
                    if mount not in vfolder_name_to_id:
                        raise BackendClientError(f"VFolder (name: {mount}) not found") from e
                    vfolder_id = vfolder_name_to_id[mount]
                extra_mount_body[str(vfolder_id)] = {
                    "mount_destination": extra_mount_map.get(mount),
                }
                if mount_type := extra_mount_options.get(mount, {}).get("type"):
                    extra_mount_body[str(vfolder_id)]["type"] = mount_type
        model_config = {
            "model": model_id_or_name,
            "model_mount_destination": model_mount_destination,
            "extra_mounts": extra_mount_body,
            "environ": envs,
            "scaling_group": scaling_group,
            "resources": resources,
            "resource_opts": resource_opts,
            "model_definition_path": model_definition_path,
        }
        if model_version:
            model_config["model_version"] = model_version
        model_config = {k: v for k, v in model_config.items() if v is not None}
        rqst = Request("POST", "/services")
        rqst_body = {
            "name": service_name,
            "replicas": initial_session_count,
            "image": image,
            "arch": architecture,
            "group": group_name,
            "domain": domain_name,
            "cluster_size": cluster_size,
            "cluster_mode": cluster_mode,
            "tag": tag,
            "startup_command": startup_command,
            "bootstrap_script": bootstrap_script,
            "owner_access_key": owner_access_key,
            "open_to_public": expose_to_public,
            "runtime_variant": runtime_variant,
            "config": model_config,
        }
        rqst_body = {k: v for k, v in rqst_body.items() if v is not None}
        rqst.set_json(rqst_body)
        async with rqst.fetch() as resp:
            body = await resp.json()
            return {
                "endpoint_id": body["endpoint_id"],
                "name": service_name,
                "replicas": initial_session_count,
                "active_route_count": 0,
                "service_endpoint": None,
                "is_public": expose_to_public,
            }

    @api_function
    @classmethod
    async def try_start(
        cls,
        image: str,
        model_id_or_name: str,
        *,
        resources: Mapping[str, str | int],
        resource_opts: Mapping[str, str | int],
        domain_name: str,
        group_name: str,
        scaling_group: str,
        service_name: str | None = None,
        model_version: str | None = None,
        _dependencies: Sequence[str] | None = None,
        model_mount_destination: str | None = None,
        envs: Mapping[str, str] | None = None,
        startup_command: str | None = None,
        cluster_size: int = 1,
        cluster_mode: Literal["single-node", "multi-node"] = "single-node",
        bootstrap_script: str | None = None,
        tag: str | None = None,
        architecture: str | None = DEFAULT_IMAGE_ARCH,
        owner_access_key: str | None = None,
        expose_to_public: bool = False,
    ) -> dict[str, Any]:
        """
        Tries to start an inference session and terminates immediately.

        :param image: The image name and tag for the infernence session.
            Example: ``python:3.6-ubuntu``.
            Check out the full list of available images in your server using (TODO:
            new API).
        :param service_name: A client-side (user-defined) identifier to distinguish the session among currently
            running sessions.
            It may be used to seamlessly reuse the session already created.
        :param mounts: The ID of model type vFolder which contains model files required
            to start inference session.
        :param model_mount_destination: Path inside the container to mount model vFolder,
            defaults to /models
        :param envs: The environment variables which always bypasses the jail policy.
        :param resources: The resource specification. (TODO: details)
        :param cluster_size: The number of containers in this compute session.
            Must be at least 1.
        :param cluster_mode: Set the clustering mode whether to use distributed
            nodes or a single node to spawn multiple containers for the new session.
        :param tag: An optional string to annotate extra information.
        :param owner: An optional access key that owns the created session. (Only
            available to administrators)
        :param expose_to_public: Visibility of API Endpoint which serves inference workload.
            If set to true, no authentication will be required to access the endpoint.

        :returns: The :class:`ComputeSession` instance.
        """
        if service_name is None:
            faker = Faker()
            service_name = f"bai-serve-{faker.user_name()}"[:SESSION_NAME_MAX_LENGTH]

        rqst = Request("POST", "/services/_/try")
        rqst.set_json({
            "name": service_name,
            "replicas": 1,
            "image": image,
            "arch": architecture,
            "group": group_name,
            "domain": domain_name,
            "cluster_size": cluster_size,
            "cluster_mode": cluster_mode,
            "tag": tag,
            "startup_command": startup_command,
            "bootstrap_script": bootstrap_script,
            "owner_access_key": owner_access_key,
            "open_to_public": expose_to_public,
            "config": {
                "model": model_id_or_name,
                "model_version": model_version,
                "model_mount_destination": model_mount_destination,
                "environ": envs,
                "scaling_group": scaling_group,
                "resources": resources,
                "resource_opts": resource_opts,
            },
        })
        async with rqst.fetch() as resp:
            body = await resp.json()
            return {
                "task_id": body["task_id"],
                "name": service_name,
            }

    def __init__(self, id: str | UUID) -> None:
        super().__init__()
        self.id = id if isinstance(id, UUID) else UUID(id)

    @api_function
    async def info(self) -> dict[str, Any]:
        rqst = Request("GET", f"/services/{self.id}")
        async with rqst.fetch() as resp:
            result: dict[str, Any] = await resp.json()

            return result

    @api_function
    async def delete(self) -> dict[str, Any]:
        rqst = Request("DELETE", f"/services/{self.id}")
        async with rqst.fetch() as resp:
            result: dict[str, Any] = await resp.json()

            return result

    @api_function
    async def sync(self) -> dict[str, Any]:
        rqst = Request("POST", f"/services/{self.id}/sync")
        async with rqst.fetch() as resp:
            result: dict[str, Any] = await resp.json()

            return result

    @api_function
    async def scale(self, to: int) -> dict[str, Any]:
        rqst = Request("POST", f"/services/{self.id}/scale")
        rqst.set_json({"to": to})
        async with rqst.fetch() as resp:
            result: dict[str, Any] = await resp.json()

            return result

    @api_function
    async def generate_api_token(self, duration: str) -> dict[str, Any]:
        rqst = Request("POST", f"/services/{self.id}/token")
        rqst.set_json({"duration": duration})
        async with rqst.fetch() as resp:
            result: dict[str, Any] = await resp.json()

            return result

    @api_function
    async def update_traffic_ratio(self, target_route_id: UUID, new_ratio: float) -> dict[str, Any]:
        rqst = Request("PUT", f"/services/{self.id}/routings/{target_route_id}")
        rqst.set_json({"traffic_ratio": new_ratio})
        async with rqst.fetch() as resp:
            result: dict[str, Any] = await resp.json()

            return result

    @api_function
    async def downscale_single_route(self, target_route_id: UUID) -> dict[str, Any]:
        rqst = Request("DELETE", f"/services/{self.id}/routings/{target_route_id}")
        async with rqst.fetch() as resp:
            result: dict[str, Any] = await resp.json()

            return result
